from sqlalchemy.sql import or_
from pipruncommon import IndexEntitiesRequest

class Operator():
    EQUAL = "eq"
    NOT_EQUAL = "ne"
    GREATER_THAN = "gt"
    LESS_THAN = "lt"
    GREATER_OR_EQUAL = 'ge'
    LESS_OR_EQUAL = 'le'
    EXISTS = "exists"
    CONTAINS = "contains"
    STARTSWITH = "startswith"

class Order():
    ASC = "Asc"
    DESC = "Desc"

def build_query_list(cls, index_entities_request:IndexEntitiesRequest, session) -> list:
    filters_list = index_entities_request.filters
    query_list = []
    for filter in filters_list:
        field = getattr(cls, filter.field)
        if(filter.operator == Operator.EQUAL):
            query_list.append(field == filter.values[0])
        elif(filter.operator == Operator.NOT_EQUAL):
            query_list.append(field.not_in(filter.values))
        elif(filter.operator == Operator.GREATER_THAN):
            query_list.append(field > filter.values[0])
        elif(filter.operator == Operator.LESS_THAN):
            query_list.append(field < filter.values[0])
        elif(filter.operator == Operator.GREATER_OR_EQUAL):
            query_list.append(field >= filter.values[0])
        elif(filter.operator == Operator.LESS_OR_EQUAL):
            query_list.append(field <= filter.values[0])
        elif(filter.operator == Operator.EXISTS):
            query_list.append(field.in_(filter.values))
        elif(filter.operator == Operator.CONTAINS):     # case insensitive
            query_list.append(or_(field.contains(i) for i in filter.values))
            # query_list.append(or_(field.like("%" + i + "%") for i in filter.values))  # this is ok
        elif(filter.operator == Operator.STARTSWITH):   # case insensitive
            query_list.append(or_(field.startswith(i) for i in filter.values))
            # query_list.append(or_(field.like(i + "%") for i in filter.values))    # this is ok
    return query_list

def query(cls, index_entities_request:IndexEntitiesRequest, session) -> list:
    query_list = build_query_list(cls, index_entities_request, session)
    order_list = index_entities_request.order
    pageSize = index_entities_request.pageSize
    skip = index_entities_request.skip
    order_by_list = []
    if order_list is not None:
        for order in order_list:
            field = getattr(cls, order.field)
            if(order.direction == Order.ASC):
                order_by_list.append(field.asc())
            if(order.direction == Order.DESC):
                order_by_list.append(field.desc())
    if skip is None:
        skip = 0
    if pageSize is None:
        result_list = session.query(cls).filter(*query_list).order_by(*order_by_list).offset(skip).all()
    else:
        result_list = session.query(cls).filter(*query_list).order_by(*order_by_list).limit(pageSize).offset(skip).all()
    return result_list

def query_count(cls, index_entities_request:IndexEntitiesRequest, session) -> int:
    skip = index_entities_request.skip
    query_list = build_query_list(cls, index_entities_request, session)
    if skip is None:
        skip = 0
    count = session.query(cls).filter(*query_list).offset(skip).count()
    return count


import os
import binascii
import re

from .base_debugger import Language
from .py_debugger import PythonDebugger
from .r_debugger import RDebugger
from .debug_meta import DebugMetadata

# VSCODE_DEVCONTAINER_INVOKE_CMD_PATTERN = "code --folder-uri vscode-remote://dev-container+{hex_encoded_devcontainer_path}{app_path}"

#TODO: add R support
SUPPORTED_DEBUGGER = [Language.PYTHON, Language.R]

class DebuggerNotSuportedError(Exception):
    def __init__(self, err="debugger not supported") -> None:
        self.err = err
        Exception.__init__(self,err)

def check_job_programming_language(code_dir:str, command:str) -> Language:
    if command:
        exe = command.lstrip(" ").split(" ")[0]
        if command.startswith("python") or exe.endswith(".py"):
            return Language.PYTHON
        if command.startswith("Rscript") or command.startswith("R "):
            return Language.R
    if code_dir is None:
        return Language.NONE
    # TODO: detect language by code_dir (guesslang is too big, consider pygments first)
    return None

def prepare_debug_info(meta: DebugMetadata):
    #1. check programming language
    language = check_job_programming_language(meta.code_dir, meta.command)
    if language == Language.NONE:
        raise DebuggerNotSuportedError(err="No code is provided in this job!")
    if language not in SUPPORTED_DEBUGGER:
        raise DebuggerNotSuportedError(err="Job programming language is not supported now!")

    envs = {"MLFLOW_TRACKING_URI":"http://host.docker.internal:5000/"} if meta.debug_dir_in_container else None
    if meta.environment_variables:
        envs = meta.environment_variables.update(envs) if envs else meta.environment_variables

    # if run in container, use debug_dir_in_container first, otherwise use debug_dir_on_host
    debug_dir = meta.debug_dir_in_container if meta.debug_dir_in_container else meta.debug_dir_on_host
    # run in container
    if not os.path.exists(debug_dir):
        os.makedirs(debug_dir)

    #prepare debug folder (prepare devcontainer.json/Dockerfile/launch.json)
    if language == Language.PYTHON:
        debugger = PythonDebugger(meta.code_dir, meta.mounts, meta.command, meta.container_image, envs)
        debugger.prepare_debug_data(debug_dir)
    elif language == Language.R:
        debugger = RDebugger(meta.code_dir, meta.mounts, meta.command, meta.container_image, envs)
        debugger.prepare_debug_data(debug_dir)

    hex_encoded_devcontainer_path = _encode_hex(meta.debug_dir_on_host)
    return f"code --folder-uri vscode-remote://dev-container+{hex_encoded_devcontainer_path}/workspaces/{meta.app_name}"

def _encode_hex(path: str):
    vscode_path = re.sub("\\s+", "", path)
    return binascii.hexlify(vscode_path.encode()).decode("ascii")


from genericpath import exists
import hashlib
import logging
from typing import Dict, Union
from pathlib import Path

import docker
from docker import DockerClient
from docker.errors import ImageNotFound, APIError
from azure.core.exceptions import AzureError

from azure.ai.ml import MLClient
# from azure.ai.ml.constants._common import AzureMLResourceType
# from azure.ai.ml.operations._environment_operations import EnvironmentOperations
from azure.ai.ml.entities import CommandComponent, Environment
from azure.ai.ml.entities._builders import BaseNode

from .environment_resolver import EnvironmentResolver
from .dockerfile_resolver import DockerfileResolver
from .constants import ENV_CURATED_IMAGE_PREFIX, MCR_CURATED_IMAGE_PREFIX, CONDA_FILE_NAME
from .logger import get_logger

AZURE_ENV_PATH="/azureml-envs/inf-conda-env"

class LocalImageBuildError(Exception):
    def __init__(self, error: Union[str, Exception]):
        err = f"Building the local image failed with error: {str(error)}"
        super().__init__(err)

class LocalImageBuilder(object):
    """
    """
    def __init__(self, ml_client: MLClient, log_path: str) -> None:
        self._environment_resolver = EnvironmentResolver()
        self._ml_client = ml_client
        self._docker_client: DockerClient = docker.from_env()
        self._logger = get_logger("image-builder", log_path=log_path, std_out=True)

    def _get_build_directory(self, pipeline_name: str, job_name: str) -> Path:
        return Path(Path.home(), ".azureml", "pipeline", pipeline_name, job_name)
    
    def build_container_image(self, pipeline_name: str, job: BaseNode):
        """build job docker container image based on environment info in job component.
        
        :return local docker container image name
        """
        if not isinstance(job._component, CommandComponent):
            raise AzureError(message="local job image build only support command component.")
        environment=job._component.environment
        if self._environment_resolver._environment_contains_cloud_artifacts(environment):
            name, version = self._environment_resolver._parse_environment_name_version(job._component.environment)
            if not name.lower().startswith(ENV_CURATED_IMAGE_PREFIX):
                raise AzureError(message="Only azureml curated environments (https://learn.microsoft.com/en-us/azure/machine-learning/resource-curated-environments) are supported in local run mode!")
            image_name = self._pull_curated_image_from_mcr(name, version)
            return image_name, None
        # if only image is provided in environment, return directly
        if isinstance(environment, Environment) and environment.image is not None and environment.conda_file is None:
            return environment.image, None

        # image is related to environment, so use environment_name:environment_version as the local image name
        image_name = self.get_image_name(environment=environment)
        
        try:
            self._docker_client.images.get(image_name)
            self._logger.info(f"image with name {image_name} already exist, skip image building...")
            return image_name, None
        except ImageNotFound:
            pass
        except APIError as e:
            self._logger.error(f"get image with name {image_name} err:{e}")
            raise LocalImageBuildError(e)
        
        download_dir = self._get_build_directory(pipeline_name, job.name)
        download_dir.mkdir(parents=True, exist_ok=True)
        # env_operation: EnvironmentOperations = None
        # if self._ml_client is not None:
        #     env_operation = self._ml_client._operation_container.all_operations.get(AzureMLResourceType.ENVIRONMENT)
        download_path = str(download_dir.resolve())
        (
            yaml_base_image_name,
            yaml_env_conda_file_path,
            yaml_env_conda_file_contents,
            downloaded_build_context,
            yaml_dockerfile,
        ) = self._environment_resolver.get_environment_artifacts(job)
        # code_directory_path = Path(job._component.base_path, job._component.code)

        if yaml_env_conda_file_contents:
            self._write_conda_file(
                conda_contents=yaml_env_conda_file_contents,
                directory_path=download_path,
                conda_file_name=CONDA_FILE_NAME,
            )
            dockerfile = DockerfileResolver(
                dockerfile=yaml_dockerfile,
                docker_base_image=yaml_base_image_name,
                docker_azureml_env_path=AZURE_ENV_PATH,
                docker_conda_file_name=CONDA_FILE_NAME,
            )
        else:
            dockerfile = DockerfileResolver(
                dockerfile=yaml_dockerfile,
                docker_base_image=yaml_base_image_name,
                docker_azureml_env_path=AZURE_ENV_PATH,
                docker_conda_file_name=None,
            )
        build_directory = downloaded_build_context if downloaded_build_context else download_path
        dockerfile.write_file(directory_path=build_directory)
        # user_environment_variables = job._component.environment_variables
        self._logger.debug(f"Start building local image: [{image_name}]")
        self._logger.debug(f"Build directory: {build_directory}")
        self._logger.debug(f"Dockerfile path: {dockerfile.local_path}")
        # docker client build container
        self._build_image(build_directory=build_directory, image_name=image_name, dockerfile_path=dockerfile.local_path)
        return image_name, str(dockerfile)
    
    def _build_image(
        self,
        build_directory: str,
        image_name: str,
        dockerfile_path: str,
    ) -> None:
        try:
            self._logger.info(f"Start building docker image {image_name} from Dockerfile: {dockerfile_path}")
            self._logger.info(f"----------------------------------------[{image_name} build start]----------------------------------------")
            for status in self._docker_client.api.build(
                path=build_directory, tag=image_name, dockerfile=dockerfile_path, pull=True, decode=True, quiet=False, buildargs={"ENV": "local"}
            ):
                if "stream" in status:
                    if "An unexpected error has occurred. Conda has prepared the above report." in status["stream"]:
                        raise LocalImageBuildError(f"Issue creating conda environment:\n{status['stream']}")
                    content:str = status["stream"]
                    f = content.strip('\n\r')
                    if len(f) > 0:
                        self._logger.info(f)
                if "error" in status:
                    self._logger.info(status["error"])
                    raise Exception(status["error"])
            self._logger.info(f"----------------------------------------[{image_name} build finish]----------------------------------------")
        except APIError as e:
            raise LocalImageBuildError(e)
        except Exception as e:
            if isinstance(e, LocalImageBuildError):
                raise
            raise LocalImageBuildError(e)

    def _pull_curated_image_from_mcr(self, name: str, version: str):
        if name.lower().startswith(ENV_CURATED_IMAGE_PREFIX):
            name = name[len(ENV_CURATED_IMAGE_PREFIX):]
        if version is None:
            version = "latest"
        image_name = MCR_CURATED_IMAGE_PREFIX + name
        image_with_exact_version = image_name + ":" + version
        # check image exist or not
        exist = self._is_image_exist(image_with_exact_version)
        if exist:
            return image_with_exact_version

        # using exact tag to pull image
        try:
            self._docker_client.images.pull(image_name, tag=version)
            return image_with_exact_version
        except APIError as e:
            self._logger.info(f"image with name {image_name} and tag {version} not found!")
            if version != "latest":
                self._logger.info(f"fall back to using latest tag for image {image_name}!")
                pass
            else:
                raise e

        # try using latest tag to pull image
        try:
            self._docker_client.images.pull(image_name)
            return f"{image_name}:latest"
        except APIError as e:
            self._logger.error(f"pull image with name {image_name} failed! err:{e}")
            raise e

    def _write_conda_file(self, conda_contents: str, directory_path: str, conda_file_name: str):
        """Writes out conda file to provided directory

        :param conda_contents: contents of conda yaml file provided by user
        :type conda_contents: str
        :param directory_path: directory on user's local system to write conda file
        :type directory_path: str
        """
        conda_file_path = f"{directory_path}/{conda_file_name}"
        p = Path(conda_file_path)
        p.write_text(conda_contents)

    def get_image_name(self, environment: Union[str, Environment]):
        # TODO: using md5 from environment content should be better to avoid conflict
        if isinstance(environment, Environment):
            return f"{environment.name.lower()}:{environment.version}"
        # for string format environment, should we respect the verstion tag?
        return get_md5_string(environment)

    def _is_image_exist(self, image_name:str) -> bool:
        try:
            self._docker_client.images.get(image_name)
            self._logger.info(f"image with name {image_name} already exist, skip image building...")
            return True
        except ImageNotFound:
            pass
        except APIError as e:
            self._logger.error(f"get image with name {image_name} err:{e}")
        return False

def get_md5_string(text):
    try:
        return hashlib.md5(text.encode("utf8")).hexdigest()
    except Exception as ex:
        raise ex



from __future__ import annotations
from typing import Union
from enum import Enum

# define input type for job/components
class InputType:
    # asset type
    URI_FILE = "uri_file"
    URI_FOLDER = "uri_folder"
    MLTABLE = "mltable"
    MLFLOW_MODEL = "mlflow_model"
    CUSTOM_MODEL = "custom_model"
    TRITON_MODEL = "triton_model"
    # parameter type
    INTEGER = "integer"
    NUMBER = "number"
    STRING = "string"
    BOOLEAN = "boolean"

# define output type for job/components
class OutputType:
    URI_FILE = "uri_file"
    URI_FOLDER = "uri_folder"
    MLTABLE = "mltable"
    MLFLOW_MODEL = "mlflow_model"
    CUSTOM_MODEL = "custom_model"
    TRITON_MODEL = "triton_model"

# define mode for job/component input/output
class InputOutputModes:
    MOUNT = "mount"
    DOWNLOAD = "download"
    UPLOAD = "upload"
    RO_MOUNT = "ro_mount"
    RW_MOUNT = "rw_mount"
    EVAL_MOUNT = "eval_mount"
    EVAL_DOWNLOAD = "eval_download"
    DIRECT = "direct"

# component input
class GeneralInput(object):
    ALLOWED_PARAMETER_TYPES = {
        "integer": (int),
        "string": (str),
        "number": (float),
        "boolean": (bool),
    }

    def __init__(self, *, name: str, type: str = InputType.URI_FILE,
        path: str = None,
        default: Union[str, int, float, bool] = None,
        min: Union[int, float] = None,
        max: Union[int, float] = None,
        description: str = None,
        optional: bool = None,
        mode: str = None,
        **kwargs):
        self.name = name
        self.type = type
        self.path = path
        self.default = default
        self.min = min
        self.max = max
        self.description = description
        self.optional = optional
        self.mode = mode

    def is_parameter_type(self) -> bool:
        return self.type in self.ALLOWED_PARAMETER_TYPES

# component output
class GeneralOutput(object):
     def __init__(self, *, name: str, type: str = OutputType.URI_FOLDER,
        path: str = None,
        description: str = None,
        mode: str = None,
        **kwargs):
        self.name = name
        self.type = type
        self.path = path
        self.description = description
        self.mode = mode

class RunInputOutputBase(object):
    def __init__(self, name:str = None, type:str = None, uri: str = None, mode:str = None, value:str = None, pipelineReference: str = None):
        self.name = name
        self.type = type
        self.uri = uri
        self.mode = mode
        self.value = value
        self.pipelineReference = pipelineReference


from typing import Dict
from .input_output import GeneralInput, GeneralOutput

class ComponentType:
    COMMAND = "CommandComponent"

class BaseComponent(object):
    def __init__(self,
        component_id:str,
        component_type:str = ComponentType.COMMAND,
        description:str = None,
        name: str = None,
        display_name :str = None,
        version:str = None,
        tags: dict = None,
        inputs :Dict[str, GeneralInput] = None,
        outputs :Dict[str, GeneralOutput] = None,
        is_deterministic: bool = False,
        raw_content: str = None,
        **kwargs
        ):
        self.component_id = component_id
        self.component_type = component_type
        self.description = description
        self.name = name
        self.display_name = display_name
        self.version = version
        self.tags = tags
        self.inputs = inputs
        self.outputs = outputs
        self.is_deterministic = is_deterministic
        self.raw_content = raw_content    

class CommandComponent(BaseComponent):
    def __init__(self,
        component_id:str,
        component_type:str = ComponentType.COMMAND,
        description:str = None,
        name: str = None,
        display_name :str = None,
        version:str = None,
        tags: dict = None,
        properties: dict = None,
        inputs :Dict[str, GeneralInput] = None,
        outputs :Dict[str, GeneralOutput] = None,
        is_deterministic: bool = False,
        command: str = None,
        code: str = None,
        distribution: str = None,
        resources: str = None,
        raw_content: str = None,
        **kwargs
        ):
        super().__init__(component_id, component_type, description, name, display_name, version, tags, inputs, outputs, is_deterministic, raw_content)
        self.command = command
        self.properties = properties
        self.code = code
        self.distribution = distribution
        self.resources = resources



# deperated
CURATED_ENVS_SUPPORTED = ["azureml-sklearn-0.24-ubuntu18.04-py37-cpu"]
ENV_CURATED_IMAGE_PREFIX = "azureml-"
MCR_CURATED_IMAGE_PREFIX = "mcr.microsoft.com/azureml/curated/"

CONDA_FILE_NAME = "conda.yml"
CONDA_ENV_NAME = "inf-conda-env"

BLOB_WASBS_REGEX = "wasbs://(.*?)@(.*?).blob.core.windows.net/(.+)"
BLOB_HTTPS_PATTERN = "https://{}.blob.core.windows.net/{}/{}"

class LocalRunMode(object):
    # so far only support COMPUTE_LOCAL_METRICS_LOCAL
    COMPUTE_LOCAL_METRICS_LOCAL = "COMPUTE_LOCAL_METRICS_LOCAL"
    COMPUTE_LOCAL_METRICS_REMOTE = "COMPUTE_LOCAL_METRICS_REMOTE"
    COMPUTE_REMOTE_METRICS_REMOTE = "COMPUTE_REMOTE_METRICS_REMOTE"

class CuratedEnvironmentDefaultRunMode(object):
    NATIVE = "NATIVE"
    CONTAINER = "CONTAINER"

class JobRunningStatus(object):
    NOT_STARTED = "NotStarted"
    RUNNING = "Running"
    FAIL = "Failed"
    SUCCESS = "Completed"
    Canceled = "Canceled"

class JobLocalRunMode(object):
    # run as native process
    NATIVE = "native"
    # directly use provided container image to run
    DOCKER = "docker"


from .graph import build_command_template, generate_runnable_command, PipelineExecutionGraph, PipelineExecutionNode
from .environment_resolver import EnvironmentResolver
from .image_builder import LocalImageBuilder, LocalImageBuildError
from .executor import run
from .constants import LocalRunMode, CuratedEnvironmentDefaultRunMode


import os
from pathlib import Path
from typing import Dict, Union, Iterable, Tuple, Optional

from azure.ai.ml.entities._assets.environment import BuildContext, Environment
from azure.ai.ml.entities._builders import BaseNode
from azure.ai.ml._utils._arm_id_utils import parse_prefixed_name_version

from azure.core.exceptions import AzureError

from ._utils import dump_yaml

class EnvironmentResolver:
    def get_environment_artifacts(
        self,
        job: BaseNode,
        # environment_operations: EnvironmentOperations,
        # download_path: str,
    ) -> Iterable[str]:
        """Validates and returns artifacts from environment specification.

        :param job: job to validate
        :type job: BaseNode entity
        :return: (base_image, conda_file_path, conda_file_contents, build_directory, dockerfile_contents) - Either base_image or build_directory should be None.
        :type return: Iterable[str]
        """
        # remote environment is not supported(besides curated environment)
        # if self._environment_contains_cloud_artifacts(environment=job._component.environment): 
        #     name, version = self._parse_environment_name_version(job._component.environment)
        #     environment_asset = environment_operations.get(name=name, version=version)
        #     if self._is_microsoft_curated_environment(environment_asset):
        #         raise AzureError(message=f"Currently can't support building curated environment images in pipeline local run. Environment: {environment_asset}")
        #     if not self._cloud_environment_is_valid(environment=environment_asset):
        #         raise AzureError(
        #             message="Cloud environment must have environment.image "
        #             + "or the environment.build.path set to work for local endpoints.",
        #         )
        #     return self._get_cloud_environment_artifacts(
        #         environment_operations=environment_operations,
        #         environment_asset=environment_asset,
        #         download_path=download_path,
        #     )
        if not self._local_environment_is_valid(environment=job._component.environment):
            raise AzureError(message=f"Local job run only support local artifacts. Job {job.name} did not contain required local artifact environment.image or environment.build.path.")
        return self._get_local_environment_artifacts(job._component.base_path, job._component.environment)

    # def _get_cloud_environment_artifacts(
    #     self,
    #     environment_operations: EnvironmentOperations,
    #     environment_asset: Environment,
    #     download_path: str,
    # ) -> Tuple[str, str, str, str]:
    #     """
    #     :return: (base_image, conda_file_path, conda_file_contents, build_directory, dockerfile_contents) - Either base_image or build_directory should be None.
    #     :type return: Iterable[str]
    #     """
    #     if environment_asset.build and environment_asset.build.path and is_url(environment_asset.build.path):
    #         environment_build_directory = download_artifact_from_storage_url(
    #             blob_url=environment_asset.build.path,
    #             destination=download_path,
    #             datastore_operation=environment_operations._datastore_operation,
    #             datastore_name=None,
    #         )
    #         dockerfile_path = Path(environment_build_directory, environment_asset.build.dockerfile_path)
    #         dockerfile_contents = dockerfile_path.read_text()
    #         return (None, None, None, environment_build_directory, dockerfile_contents)
    #     conda_file_contents = (
    #         dump_yaml(environment_asset.conda_file) if environment_asset.conda_file else None
    #     )
    #     return (environment_asset.image, environment_asset.id, conda_file_contents, None, None)

    def _get_local_environment_artifacts(self, base_path: str, environment: Environment):
        """
        :return: (base_image, conda_file_path, conda_file_contents, build_directory, dockerfile_contents) - Either base_image or build_directory should be None.
        :type return: Iterable[str]
        """
        if environment.image:
            conda_file_contents = dump_yaml(environment.conda_file)
            return (environment.image, environment._conda_file_path, conda_file_contents, None, None)

        dockerfile_contents = None
        if environment.build:
            if base_path is None:
                base_path = os.getcwd()
            absolute_build_directory = Path(base_path, environment.build.path).resolve()
            docker_file=environment.build.dockerfile_path
            if docker_file is None:
                # default docker file name
                docker_file = "Dockerfile"
            absolute_dockerfile_path = Path(absolute_build_directory, docker_file).resolve()
            dockerfile_contents = absolute_dockerfile_path.read_text()
            return (None, None, None, str(absolute_build_directory), dockerfile_contents)

    def _environment_contains_cloud_artifacts(self, environment: Union[str, Environment]) -> bool:
        return isinstance(environment, str)
    
    def _is_microsoft_curated_environment(self, environment: Environment) -> bool:
        created_info = environment.creation_context
        return created_info.created_by == "Microsoft"
    
    def _local_environment_is_valid(self, environment: Union[str, Environment]) -> bool:
        return isinstance(environment, Environment) and (
            environment.image
            or (
                environment.build is not None
                and isinstance(environment.build, BuildContext)
                and self._local_build_context_is_valid(environment.build)
            )
        )

    def _local_build_context_is_valid(self, build_context: BuildContext):
        return build_context.path is not None

    def _cloud_environment_is_valid(self, environment: Environment):
        return isinstance(environment, Environment) and (
            environment.image or (environment.build and environment.build.path)
        )
    
    def _parse_environment_name_version(self, environment: str) -> Tuple[str, Optional[str]]:
        name, version = parse_prefixed_name_version(environment)
        if version is None and "@" in name:
            at_splits = name.rsplit("@", 1)
            name=at_splits[0]
            version = at_splits[1]
        return name, version




import os
from pathlib import Path
from sqlalchemy import create_engine
from sqlalchemy.orm import sessionmaker
from sqlalchemy.ext.declarative import declarative_base

DEFAULT_DB_PATH = os.path.join(os.path.sep, str(Path.home()), ".azureml", "piprun", "localrun.db")

def db_setup(default_db_path: str = None, table_list: list = None):
    Base = declarative_base()
    db_path = os.environ.get('AML_LOCAL_RUN_DB_PATH')
    if db_path is None:
        db_path = default_db_path
    if db_path is None:
        db_path = DEFAULT_DB_PATH
    engine = create_engine(f'sqlite:///{db_path}', echo = False)
    Base.metadata.create_all(
        engine,
        tables = table_list,
        checkfirst = True
    )
    Session = sessionmaker(bind = engine)
    return Session

def drop_tables(default_db_path: str = None, table_list: list = None):
    Base = declarative_base()
    db_path = os.environ.get('AML_LOCAL_RUN_DB_PATH')
    if db_path is None:
        db_path = default_db_path
    if db_path is None:
        db_path = DEFAULT_DB_PATH
    engine = create_engine(f'sqlite:///{db_path}', echo = False)
    Base.metadata.drop_all(engine, tables = table_list)


from .component_metadata import ComponentMetadata
from .experiment_metadata import ExperimentMetadata
from .run_definition import RunDefinition
from .run_history import RunHistory

all_tables = [
    ComponentMetadata.__table__,
    ExperimentMetadata.__table__,
    RunDefinition.__table__,
    RunHistory.__table__,
]


from __future__ import annotations
from abc import abstractmethod
from typing import Dict, Union
import uuid
import hashlib
from azure.ai.ml.entities import CommandComponent, PipelineJob
from azure.ai.ml import Input,Output
from azure.ai.ml.entities._builders import BaseNode
from azure.ai.ml.entities._assets import Environment
from azure.ai.ml.entities._job.pipeline._io import NodeInput, NodeOutput, PipelineInput, PipelineOutput
import re
from string import Template
import os

from pipruncommon import RunType

from .constants import JobRunningStatus, JobLocalRunMode, CuratedEnvironmentDefaultRunMode

class BaseExecutionNode(object):
    """
    """
    def __init__(
        self,
        output_root_dir: str,
        job: BaseNode,
        curated_env_default_run_mode = CuratedEnvironmentDefaultRunMode.CONTAINER,
        **kwargs,
        ):
        self._status = JobRunningStatus.NOT_STARTED
        self._job = job
        self._command: str = job.command
        # create output sub dir for job
        self._outputdir = os.path.join(output_root_dir, job.name)
        if not os.path.exists(self._outputdir):
            os.makedirs(self._outputdir)
        self._curated_env_default_run_mode = curated_env_default_run_mode
    
    @abstractmethod
    def inputs() -> Dict[str, Union[Input, PipelineInput, Output, NodeOutput]]:
        pass

    @abstractmethod
    def outputs() -> Dict[str, Union[Output, NodeOutput]]:
        pass

    def status(self) -> str:
        return self._status

    def mark_run_result(self, success: bool):
        if success:
            self._status = JobRunningStatus.SUCCESS
        else:
            self._status = JobRunningStatus.FAIL

    def is_finish(self) -> bool:
        return self._status == JobRunningStatus.SUCCESS or self._status == JobRunningStatus.FAIL

    @abstractmethod
    def is_ready(selft) -> bool:
        pass

    def get_run_mode(self) -> str:
        if "image" in self._job.tags.keys():
            return JobLocalRunMode.DOCKER
        # need to build container for local environment
        if isinstance(self._job._component, CommandComponent):
            env = self._job._component.environment
            if isinstance(env, Environment) and (env.image or (env.build and env.build.path)):
                return JobLocalRunMode.DOCKER
            if isinstance(env, str) and self._curated_env_default_run_mode == CuratedEnvironmentDefaultRunMode.CONTAINER:
                return JobLocalRunMode.DOCKER
        return JobLocalRunMode.NATIVE

    def command(self) -> str:
        return self._command

_ALLOWED_TYPES = {
        "integer": (int),
        "string": (str),
        "number": (float),
        "boolean": (bool),
    }

class PipelineExecutionNode(object):
    """
    """
    def __init__(
        self,
        output_root_dir: str,
        job: BaseNode,
        pipeline_inputs: Dict[str, NodeInput],
        pipeline_outputs: Dict[str, NodeOutput],
        curated_env_default_run_mode = CuratedEnvironmentDefaultRunMode.CONTAINER,
        **kwargs,
        ):
        self._status = JobRunningStatus.NOT_STARTED
        self._job = job
        self._run_id = str(uuid.uuid1())
        self._node_id = str(uuid.uuid1())
        self._parent_run_id = kwargs.get("parent_run_id", None)
        self._root_run_id = kwargs.get("root_run_id", None)
        self._experiment_name = kwargs.get("experiment_name", None)
        self._experiment_id = kwargs.get("experiment_id", None)
        self._run_type = RunType.COMMAND_JOB
        #if isinstance(job, Command):
        self._command: str = job.command
        component_str = job._component._to_yaml()
        self._component_id = hashlib.md5(component_str.encode('utf-8')).hexdigest()
        job._component._id = self._component_id
        self._dependencies: Dict[str, PipelineExecutionNode] = {}
        # create output sub dir for job
        self._outputdir = os.path.join(output_root_dir, job.name, "outputs")
        if not os.path.exists(self._outputdir):
            os.makedirs(self._outputdir)
        # prepare input path for those data which we need to download from remote
        self._inputs_temp_dir = os.path.join(output_root_dir, "inputs")
        # self._log_path = os.path.join(os.sep, output_root_dir, f"{job.name}_run.log")  # to be refined for mpi job
        self._log_path = os.path.join(os.sep, output_root_dir, job.name, "run.log") 
        self._pipeline_inputs = pipeline_inputs
        self._pipeline_outputs = pipeline_outputs
        self._curated_env_default_run_mode = curated_env_default_run_mode
        # Dict[str, Union[int, bool, float, str, Input, JobInput]]
        self._inputs: Dict[str, NodeInput] = {}
        self._outputs: Dict[str, NodeOutput] = {}
        # iterate node output(PipelineOutput) dic
        for key, output in job.outputs.items():
            # get job output with type `Output`
            job_output = output._to_job_output()
            if job_output is None:
                v = PipelineOutput(name=key, data=Output(path=self._outputdir), meta=output._meta, owner=job)
                self._outputs[key] = v
                continue
            d = job_output.path
            # try to find pipeline output or create new if not defined
            if isinstance(d, str) and d.startswith("$"):
                d = d.strip("${}")
                sections = d.split(".")
                if len(sections) != 3:
                    raise Exception(f"Not supported pipeline job output: {d}")
                output_key = sections[2]
                if output_key in pipeline_outputs.keys():
                    v = pipeline_outputs[output_key]
                    meta = job._component.outputs[key]
                    if v._data is None:
                        joboutput = Output(path=self._outputdir)
                        # joboutput = Output(path=output_root_dir)
                        v = PipelineOutput(name=key, data=joboutput, meta=meta, owner=job)
                    else:
                        if v.path is None:
                            v.path = ""
                        if not os.path.isabs(v.path):
                            v.path = os.path.join(self._outputdir, v.path.strip("./\\"))
                        # TODO: move such logic to a common pre-run step
                        if not os.path.exists(v.path):
                            os.makedirs(v.path)
                        v = PipelineOutput(name=key, data=v._data, meta=meta, owner=job)
                    self._outputs[key] = v
                else:
                    raise Exception(f"job output with key {output_key} not defined in pipeline")
            else:
                if output.path is None:
                    output.path = self._outputdir
                self._outputs[key] = output

    def set_dependencies(self, dependencies: Dict[str, PipelineExecutionNode]):
        self._dependencies = dependencies
        # regenerate input
        for key, input in self._job.inputs.items():
            job_input = input._to_job_input()
            if not isinstance(job_input, Input):
                self._inputs[key] = input
                continue
            input_data = job_input.path
            meta = self._job._component.inputs[key]
            if isinstance(input_data, str) and input_data.startswith("$"):
                input_data = input_data.strip("${}")
                sections = input_data.split(".")
                if len(sections) == 3:
                    # pipeline input
                    input_key = sections[2]
                    if input_key in self._pipeline_inputs.keys():
                        v = self._pipeline_inputs[input_key]
                        if v is None:
                            raise Exception(f"job input {input_key} is found empty in pipeline")
                        # self._inputs[key] = PipelineInput(name=input_key, meta=meta, data=v._data)
                        self._inputs[key] = NodeInput(name=key, meta=meta, data=v, owner=self._job)
                    else:
                        raise Exception(f"job input with key {input_key} not defined in pipeline")
                elif len(sections) == 5:
                    # other job input
                    src_job = sections[2]
                    job_output_key = sections[4]
                    if src_job not in dependencies.keys():
                        raise Exception(f"dependency job with name {src_job} not found")
                    src_node = dependencies[src_job]
                    if job_output_key not in src_node.outputs().keys():
                        raise Exception(f"output name {job_output_key} not found in dependency job {src_job}")
                    self._inputs[key] = NodeInput(name=key, meta=meta, data=src_node.outputs()[job_output_key], owner=self._job)
                else:
                    raise Exception(f"Not supported pipeline job input: {input_data}")
            else:
                 self._inputs[key] = input
        # generate final runnable command
        self._command_template = build_command_template(self._command, self._inputs, self._outputs)
        self._command = generate_runnable_command(self._command, self._inputs, self._outputs)

    def status(self) -> str:
        return self._status

    def mark_run_result(self, success: bool):
        if success:
            self._status = JobRunningStatus.SUCCESS
        else:
            self._status = JobRunningStatus.FAIL

    def is_ready(self) -> bool:
        ready = True
        for _, node in self._dependencies.items():
            if node.status() != JobRunningStatus.SUCCESS:
                ready = False
                break
        return ready
    
    def is_finish(self) -> bool:
        return self._status == JobRunningStatus.SUCCESS or self._status == JobRunningStatus.FAIL
    
    def get_run_mode(self) -> str:
        if "image" in self._job.tags.keys():
            return JobLocalRunMode.DOCKER
        # need to build container for local environment
        if isinstance(self._job._component, CommandComponent):
            env = self._job._component.environment
            if isinstance(env, Environment) and (env.image or (env.build and env.build.path)):
                return JobLocalRunMode.DOCKER
            if isinstance(env, str) and self._curated_env_default_run_mode == CuratedEnvironmentDefaultRunMode.CONTAINER:
                return JobLocalRunMode.DOCKER
        return JobLocalRunMode.NATIVE

    def inputs(self) -> Dict[str, NodeInput]:
        return self._inputs

    def command(self) -> str:
        return self._command

    def outputs(self) -> Dict[str, NodeOutput]:
        return self._outputs

def generate_runnable_command(command: str, inputs: Dict[str, NodeInput], outputs: Dict[str, NodeOutput]) -> str:
    command_template = build_command_template(command, inputs, outputs)
    template_params = {}
    for k, v in inputs.items():
        key = f"inputs_{k}"
        has_path = False
        data = None
        if isinstance(v._data, PipelineInput):
            if isinstance(v._data._data, Input):
                has_path = True
            else:
                data = v._data._data
        elif isinstance(v._data, (Input, Output, NodeOutput)):
            has_path = True
        else:
            data = v._data

        if has_path:
            template_params[key] = v._data.path
        else:
            template_params[key] = data
    for k, v in outputs.items():
        key = f"outputs_{k}"
        if isinstance(v._data, Output):
            template_params[key] = v._data.path
        else:
            template_params[key] = v._data
    return command_template.substitute(template_params)

def build_command_template(command: str, inputs: Dict[str, NodeInput], outputs: Dict[str, NodeOutput]) -> Template:
     # generate runnable command
    optinal_params = re.findall("\[.*?\]", command)
    if len(optinal_params) != 0:
        must_parts = re.split("\[.*?\]", command)
        # check whether optional param has value
        keep = []
        for p in must_parts:
            keep.append(p.strip(" "))
        for p in optinal_params:
            s = re.search("\{\{(.*)\}\}", p)
            param = s.group(1)
            param_parts = param.split(".")
            if len(param_parts) != 2:
                raise Exception(f"command parameter in wrong format: {param}")
            if param_parts[0] == "inputs":
                if param_parts[1] in inputs.keys():
                    keep.append(p.strip("[]"))
            elif param_parts[0] == "outputs":
                # this may not need
                if param_parts[1] in outputs.keys():
                    keep.append(p.strip("[]"))
            else:
                raise Exception(f"unknown param type: {param}")
        command = " ".join(keep)    
    command = command.replace("{{", "").replace("}}","").replace("inputs.", "inputs_").replace("outputs.", "outputs_")
    return Template(command)


class PipelineExecutionGraph(object):
    """
    """
    def __init__(
        self,
        output_root_dir: str,
        pipeline_job: PipelineJob,
        experiment_name:str,
        curated_env_default_run_mode = CuratedEnvironmentDefaultRunMode.CONTAINER,
        **kwargs,
        ):
        self._name = pipeline_job.name
        self._run_mode = curated_env_default_run_mode
        self._job = pipeline_job
        self._status = JobRunningStatus.NOT_STARTED
        self._nodes: Dict[str, PipelineExecutionNode] = {}
        self._dependencies: Dict[str, list] = {}
        self._run_id = kwargs.get("run_id", str(uuid.uuid1()))
        self._node_id = str(uuid.uuid1())
        self._parent_run_id = kwargs.get("parent_run_id", None)
        self._root_run_id = kwargs.get("root_run_id", self._run_id)
        self._run_type = RunType.PIPELINE_JOB
        self._output_root_dir = os.path.join(os.path.sep, output_root_dir, self._run_id)
        self._experiment_name = experiment_name
        self._experiment_id = str(uuid.uuid3(uuid.NAMESPACE_DNS, experiment_name))
        if not os.path.exists(self._output_root_dir):
            os.makedirs(self._output_root_dir)
        # pre-process pipeline inputs
        self._pipeline_inputs : Dict[str, Input] = {}
        for name, input in pipeline_job.inputs.items():
            # why this is needed?
            if hasattr(input._data, "path") and "://" not in input.path:
                input.path = os.path.abspath(input.path)
            self._pipeline_inputs[name] = input
        # create run history for pipeline run
        self._log_path = kwargs.get("log_path")
        for name, job in pipeline_job.jobs.items():
            # print(job._component._to_yaml())
            node = PipelineExecutionNode(
                output_root_dir=self._output_root_dir, 
                job=job, 
                pipeline_inputs=self._pipeline_inputs, 
                pipeline_outputs=pipeline_job.outputs, 
                curated_env_default_run_mode=curated_env_default_run_mode, 
                parent_run_id=self._run_id, 
                root_run_id=self._root_run_id,
                experiment_name=experiment_name,
                experiment_id=self._experiment_id
                )
            self._nodes[name] = node
            dependencies = self._get_dependency_jobs(job)
            self._dependencies[name] = dependencies
        for name, deps in self._dependencies.items():
            dep_nodes: Dict[str, PipelineExecutionNode] = {}
            for dp in deps:
                dep_nodes[dp] = self._nodes[dp]
            self._nodes[name].set_dependencies(dep_nodes)

    def _get_dependency_jobs(self, job: BaseNode) -> list:
        res = []
        for n, input in job.inputs.items():
            if isinstance(input._data, (Output, NodeOutput)):
                res.append(input._data._owner.name)
        return res

    def get_ready_nodes(self) -> Dict[str, PipelineExecutionNode]:
        read_nodes: Dict[str, PipelineExecutionNode] = {}
        for name, node in self._nodes.items():
            if node.is_ready() and not node.is_finish():
                read_nodes[name] = node
        return read_nodes
    
    def is_finish(self) -> bool:
        for name, node in self._nodes.items():
            if node.status() != JobRunningStatus.SUCCESS and node.status() != JobRunningStatus.FAIL:
                return False
        return True

def main():
    command = "python train.py  --training_data ${{inputs.training_data}}  [--max_epochs ${{inputs.max_epochs}}]  --learning_rate ${{inputs.learning_rate}}  [--learning_rate_schedule ${{inputs.learning_rate_schedule}}]  --model_output ${{outputs.model_output}}"
    inputs: Dict[str, Input] = {}
    outputs: Dict[str, Output] = {}
    inputs["training_data"] = Input(name="training_data", meta=None, data=PipelineInput(name="ttt", meta=None, data=Input(path="/train/data")))
    inputs["learning_rate"] =Input(name="learning_rate", meta=None, data=0.5)
    inputs["max_epochs"] = Input(name="max_epochs", meta=None, data=10)
    inputs["learning_rate_schedule"] = Input(name="learning_rate_schedule", meta=None, data="time-based")
    outputs["model_output"] = Output(name= "model_output", meta=None, data=Output(path="/train/output"))
    cmd = generate_runnable_command(command=command, inputs=inputs, outputs=outputs)
    print(cmd)

if __name__ == "__main__":
    main()


import yaml
from collections import OrderedDict
from os import PathLike
from typing import Union
from urllib.parse import urlparse
import requests
import shutil
import urllib3
urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)

def dump_yaml(*args, **kwargs):
    """A thin wrapper over yaml.dump which forces `OrderedDict`s to be
    serialized as mappings

    Otherwise behaves identically to yaml.dump
    """

    class OrderedDumper(yaml.Dumper):
        """A modified yaml serializer that forces pyyaml to represent
        an OrderedDict as a mapping instead of a sequence.
        """

        pass

    OrderedDumper.add_representer(OrderedDict, yaml.representer.SafeRepresenter.represent_dict)

    return yaml.dump(*args, Dumper=OrderedDumper, **kwargs)

def is_url(value: Union[PathLike, str]) -> bool:
    try:
        result = urlparse(str(value))
        return all([result.scheme, result.netloc])
    except ValueError:
        return False

def server_ok(url):
    # exception block
    try:
        response = requests.get(url)
         
        # check the status code
        if response.status_code == 200:
            return True
        else:
            return False
    except requests.ConnectionError as e:
        return False

def download(url, target) -> bool:
    try:
        r = requests.get(url, verify=False, stream=True)
        r.raw.decode_content = True
        with open(target, 'wb') as f:
            shutil.copyfileobj(r.raw, f)
        return True
    except:
        return False
    





from .db_query import query, Operator
from .experiment_metadata import ExperimentMetadata, ExperimentMetadataDBOperation
from .run_definition import RunDefinition, RunDefinitionDBOperation
from .run_history import RunHistory, RunHistoryDBOperation
from .component_metadata import ComponentMetadata, ComponentMetadataDBOperation
from .all_tables import all_tables


from sqlalchemy import Column, String
from sqlalchemy.ext.declarative import declarative_base

Base = declarative_base()

from .db_query import query

class RunDefinition(Base):
    __tablename__ = 'run_definition'
    run_id = Column(String, primary_key = True)
    run_definition = Column(String)
    def __repr__(self):
        return "RunDefinition(run_id=" + str(self.run_id) + ", run_definition=" + str(self.run_definition) + ")"
    def __eq__(self, other):
        return self.run_id == other.run_id and self.run_definition == other.run_definition

class RunDefinitionDBOperation:
    def __init__(self, Session):
        self.Session = Session

    def select(self, index_entities_request) -> list:
        with self.Session(expire_on_commit=False) as session:
            return query(RunDefinition, index_entities_request, session)

    def insert(self, insert_data):
        data = insert_data
        with self.Session(expire_on_commit=False) as session:
            session.add(insert_data)
            session.commit()
        return data

    def update(self, update_data):
        with self.Session(expire_on_commit=False) as session:
            ret = session.merge(update_data)
            session.commit()
            return ret
        
    def delete(self, delete_data_run_id):
        with self.Session(expire_on_commit=False) as session:
            session.query(RunDefinition).filter(RunDefinition.run_id == delete_data_run_id).delete()
            session.commit()
            return True


from marshmallow import Schema, fields, post_load, EXCLUDE

class IndexEntitiesRequestFilter ():
    def __init__(
        self,
        field: str,
        operator: str,
        values: list,
        ):
        self.field = field
        self.operator = operator
        self.values = values

class IndexEntitiesRequestFilterSchema(Schema):
    field = fields.Str()
    operator = fields.Str()
    # values = fields.List(fields.Str(allow_none = True), many = True)
    values = fields.List(fields.Raw(allow_none = True), many = True)

    @post_load
    def make_user(self, data, **kwargs):
        return IndexEntitiesRequestFilter(**data)

class IndexEntitiesRequestOrder():
    def __init__(
        self,
        field: str,
        direction: str,
        ):
        self.field = field
        self.direction  = direction   

class IndexEntitiesRequestOrderSchema(Schema):
    field = fields.Str()
    direction = fields.Str()

    @post_load
    def make_user(self, data, **kwargs):
        return IndexEntitiesRequestOrder(**data)  

class IndexEntitiesRequest():
    def __init__(
        self,
        filters: list = None,
        order: list = None,
        pageSize: int = None,
        skip: int = None,
        includeTotalResultCount: bool = None,
        ):
        self.filters = filters
        self.order = order
        self.pageSize = pageSize
        self.skip = skip
        self.includeTotalResultCount = includeTotalResultCount 

class IndexEntitiesRequestSchema(Schema):
    class Meta:
        unknown = EXCLUDE
    filters = fields.Nested(IndexEntitiesRequestFilterSchema, many = True) # if field nested object is a list, we should make many = True
    order = fields.Nested(IndexEntitiesRequestOrderSchema, many = True)
    pageSize = fields.Int()
    skip = fields.Int()
    includeTotalResultCount = fields.Boolean()

    @post_load      # Register a method to invoke after deserializing an object.
    def make_user(self, data, **kwargs):
        return IndexEntitiesRequest(**data)


from __future__ import annotations
import jsonpickle
from sqlalchemy import JSON, Column, Integer, PickleType, String, DateTime, TIMESTAMP, Interval, select
from sqlalchemy.sql import func
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import deferred
from .experiment_metadata import ExperimentMetadata
from .db_query import query, query_count
Base = declarative_base()

class RunHistory(Base):
    __tablename__ = 'run_history'
    run_id = Column(String, primary_key = True)
    run_number = Column(Integer)
    display_name = Column(String)   
    run_type = Column(String)
    job_type = Column(String)
    run_mode = Column(String)
    created_by = Column(String)
    created_time = Column(DateTime)
    ready_time = Column(DateTime)
    start_time = Column(DateTime)
    finish_time = Column(DateTime)
    duration = Column(Interval)
    status = Column(String)
    total_steps = Column(Integer)
    inputs = Column(PickleType)  #RunInput
    outputs = Column(PickleType)  #RunOutput
    parameters = Column(PickleType) #RunInput
    container_id = Column(String)
    log_file = Column(String)
    run_command = Column(String, comment="real command run in container")
    mounts = Column(JSON, comment="job run container mounts info")
    code_dir=Column(String, comment="job code dir on the node")
    reuse_run_id = Column(String)
    job_md5 = Column(String)
    experiment_id = Column(String)
    experiment_name = deferred(select([ExperimentMetadata.experiment_name]).where(ExperimentMetadata.experiment_id == experiment_id).scalar_subquery())
    parent_runid = Column(String)
    root_runid = Column(String)
    node_id = Column(String)
    run_uuid = Column(String)
    parent_run_uuid = Column(String)
    root_run_uuid = Column(String)
    component_id = Column(String)
    component_base_path = Column(String)
    run_root_dir = Column(String)
    container_image = Column(String)
    tags = Column(JSON)
    jobs = Column(JSON, comment="pipeline jobs info, only used for pipelinejob")
    graph = Column(PickleType)
    last_modified_time = Column(TIMESTAMP, server_default=func.now(), onupdate=func.current_timestamp())

    def __repr__(self):
        return ("RunHistory(run_id=" + str(self.run_id) 
        + ", run_number=" + str(self.run_number) 
        + ", display_name=" + str(self.display_name) 
        + ", run_type=" + str(self.run_type) 
        + ", job_type=" + str(self.job_type)
        +", run_mode=" + str(self.run_mode)
        +", created_by=" + str(self.created_by)
        +", created_time=" + str(self.created_time)
        +", ready_time=" + str(self.ready_time)
        +", start_time=" + str(self.start_time)
        +", finish_time=" + str(self.finish_time)
        +", duration=" + str(self.duration)
        +", status=" + str(self.status)
        +", total_steps=" + str(self.total_steps)
        +", inputs=" + jsonpickle.encode(self.inputs, unpicklable = False)
        +", outputs=" + jsonpickle.encode(self.outputs, unpicklable = False)
        +", parameters=" + jsonpickle.encode(self.parameters, unpicklable = False)
        +", container_id=" + str(self.container_id)
        +", log_file=" + str(self.log_file)
        +", run_command=" + str(self.run_command)
        +", reuse_run_id=" + str(self.reuse_run_id)
        +", job_md5=" + str(self.job_md5)
        +", experiment_id=" + str(self.experiment_id)
        +", parent_runid=" + str(self.parent_runid)
        +", root_runid=" + str(self.root_runid)
        +", node_id=" + str(self.node_id)
        +", run_uuid=" + str(self.run_uuid)
        +", parent_run_uuid=" + str(self.parent_run_uuid)
        +", root_run_uuid=" + str(self.root_run_uuid)
        +", component_id=" + str(self.component_id)
        +", component_base_path=" + str(self.component_base_path)
        +", run_root_dir=" + str(self.run_root_dir)
        +", container_image=" + str(self.container_image)
        +", tags=" + str(self.tags)
        +", jobs=" + str(self.jobs)
        +", graph=" + jsonpickle.encode(self.graph, unpicklable = False) + ")"
        )
    
    def __eq__(self, other: RunHistory):
        return {
            self.run_id == other.run_id
            and self.run_number == other.run_number 
            and self.display_name == other.display_name 
            and self.run_type == other.run_type 
            and self.job_type == other.job_type 
            and self.run_mode == other.run_mode
            and self.created_by == other.created_by
            and self.created_time == other.created_time
            and self.ready_time == other.ready_time
            and self.start_time == other.start_time
            and self.finish_time == other.finish_time
            and self.duration == other.duration
            and self.status == other.status
            and self.total_steps == other.total_steps
            and jsonpickle.encode(self.inputs) == jsonpickle.encode(other.inputs)
            and jsonpickle.encode(self.outputs) == jsonpickle.encode(other.outputs)
            and jsonpickle.encode(self.parameters) == jsonpickle.encode(other.parameters)
            and self.container_id == other.container_id
            and self.log_file == other.log_file
            and self.run_command == other.run_command
            and self.reuse_run_id == other.reuse_run_id
            and self.job_md5 == other.job_md5
            and self.experiment_id == other.experiment_id
            and self.parent_runid == other.parent_runid
            and self.root_runid == other.root_runid
            and self.node_id == other.node_id
            and self.run_uuid == other.run_uuid
            and self.parent_run_uuid == other.parent_run_uuid
            and self.root_run_uuid == other.root_run_uuid
            and self.component_id == other.component_id
            and self.component_base_path == other.component_base_path
            and self.run_root_dir == other.run_root_dir
            and self.container_image == other.container_image
            and self.tags == other.tags
            and self.jobs == other.jobs
            and jsonpickle.encode(self.graph) == jsonpickle.encode(other.graph)
        }

class RunHistoryDBOperation:
    def __init__(self, Session):
        self.Session = Session

    def count(self, index_entities_request) -> int:
        with self.Session(expire_on_commit=False) as session:
            return query_count(RunHistory, index_entities_request, session)

    def select(self, index_entities_request) -> list:
        with self.Session(expire_on_commit=False) as session:
            return query(RunHistory, index_entities_request, session)
    
    def get(self, run_id) -> RunHistory:
        with self.Session(expire_on_commit=False) as session:
            return session.query(RunHistory).filter(RunHistory.run_id == run_id).first()

    def insert(self, insert_data: RunHistory):
        with self.Session(expire_on_commit=False) as session:
            record = session.query(func.max(RunHistory.run_number)).first()
            if record is not None and record[0] is not None:
                insert_data.run_number = record[0] + 1
            session.add(insert_data)
            session.commit()
        return insert_data

    def update(self, update_data):
        with self.Session(expire_on_commit=False) as session:
            ret = session.merge(update_data)
            session.commit()
            return ret

    def delete(self, delete_data_run_id):
        with self.Session(expire_on_commit=False) as session:
            session.query(RunHistory).filter(RunHistory.run_id == delete_data_run_id).delete()
            session.commit()
            return True


from __future__ import annotations

import jsonpickle

from sqlalchemy import Column, String, Boolean, PickleType, DateTime, TIMESTAMP
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.sql import func

from .db_query import query

from pipruncommon import IndexEntitiesRequest

Base = declarative_base()

class ComponentMetadata(Base):
    __tablename__ = 'component_metadata'
    component_id = Column(String, primary_key = True)
    component_type = Column(String)
    created_time = Column(DateTime)
    last_modified_time = Column(TIMESTAMP, server_default=func.now(), onupdate=func.current_timestamp())
    component = Column(PickleType)
    docker_file_content = Column(String)

    def __repr__(self):
        return ("ComponentMetadata(component_id=" + str(self.component_id) 
        + ", component_type=" + str(self.component_type)
        + ", created_time=" + str(self.created_time)
        + ", component=" + jsonpickle.encode(self.component, unpicklable = False)
        + ", docker_file_content=" + str(self.docker_file_content) + ")"
        )

    def __eq__(self, other: ComponentMetadata):
        return (self.component_id == other.component_id 
            and self.component_type == other.component_type 
            and self.created_time == other.created_time
            and jsonpickle.encode(self.component) == jsonpickle.encode(other.component) 
            and self.docker_file_content == other.docker_file_content)

class ComponentMetadataDBOperation:
    def __init__(self, Session):
        self.Session = Session

    def select(self, index_entities_request: IndexEntitiesRequest) -> list:
        with self.Session(expire_on_commit=False) as session:
            return query(ComponentMetadata, index_entities_request, session)

    def insert(self, insert_data: ComponentMetadata):
        data = insert_data
        with self.Session(expire_on_commit=False) as session:
            session.add(insert_data)
            session.commit()
        return data

    def get(self, component_id) -> ComponentMetadata:
        with self.Session(expire_on_commit=False) as session:
            return session.query(ComponentMetadata).filter(ComponentMetadata.component_id == component_id).first()

    def update(self, update_data:ComponentMetadata):
        with self.Session(expire_on_commit=False) as session:
            ret = session.merge(update_data)
            session.commit()
            return ret

    def delete(self, delete_component_id):
        with self.Session(expire_on_commit=False) as session:
            session.query(ComponentMetadata).filter(ComponentMetadata.component_id == delete_component_id).delete()
            session.commit()
            return True


import os

class DebugMetadata(object):
    def __init__(self, 
        code_dir: str = None, 
        mounts:dict = None, 
        command:str = None, 
        image:str = None, 
        environment_variables:dict = None,
        debug_dir_in_container:str = None,
        debug_dir_on_host:str = None
    ) -> None:
        self.code_dir = code_dir
        self.mounts = mounts
        self.command = command
        self.container_image = image
        self.environment_variables = environment_variables
        self.debug_dir_in_container = debug_dir_in_container
        self.debug_dir_on_host = debug_dir_on_host
        debug_dir = debug_dir_in_container if debug_dir_in_container else debug_dir_on_host
        self.app_name = os.path.basename(debug_dir)


from __future__ import annotations

class Graph():
    def __init__(
            self, 
            datasetNodes: list = None, 
            edges: list = None,
        ):
        self.datasetNodes = datasetNodes
        self.edges = edges

class GraphEdge(object):
    def __init__(
        self, 
        sourceOutputPort: PortInfo = None, 
        destinationInputPort: PortInfo = None, 
        ):
        self.sourceOutputPort = sourceOutputPort
        self.destinationInputPort = destinationInputPort

class PortInfo(object):
    def __init__(
        self, 
        nodeId: str = None, 
        portName: str = None, 
        graphPortName: str = None, 
        isParameter: bool = False,
        ):
        self.nodeId = nodeId
        self.portName = portName
        self.graphPortName = graphPortName
        self.isParameter = isParameter 

    def __eq__(self, other: PortInfo):
        return {
            self.nodeId == other.nodeId
            and self.portName == other.portName 
            and self.graphPortName == other.graphPortName 
            and self.isParameter == other.isParameter 
        }

class GraphDatasetNode(object):
    def __init__(
        self, 
        id: str = None, 
        dataSetDefinition: DataSetDefinition = None, 
        ):
        self.id = id
        self.dataSetDefinition  = dataSetDefinition 

class DataSetDefinition(object):
    def __init__(
        self, 
        parameterName: str = None, 
        dataTypeShortName: str = None, 
        ):
        self.parameterName = parameterName
        self.dataTypeShortName = dataTypeShortName


# ---------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ---------------------------------------------------------
#
# This file contains devcontainer.json properties as Python classes.
# Reference: https://code.visualstudio.com/docs/remote/devcontainerjson-reference


from typing import Optional


class Image(object):
    """Python object representation of devcontainer image property."""

    def __init__(self, image: str):
        self._image = image

    def to_dict(self) -> dict:
        return {"image": self._image}


class Build(object):
    """Python object representation of devcontainer build.dockerfile property."""

    def __init__(
        self,
        dockerfile_path: str,
        build_context: Optional[str] = None,
        args: Optional[dict] = None,
        target: Optional[str] = None,
    ):
        self._dockerfile_path = dockerfile_path
        self._build_context = build_context
        self._args = args
        self._target = target

    def to_dict(self) -> dict:
        build = {
            "build": {
                "dockerfile": self._dockerfile_path,
            }
        }
        if self._build_context:
            build["build"]["context"] = self._build_context
        if self._args:
            build["build"]["args"] = self._args
        if self._target:
            build["build"]["target"] = self._target
        return build


class ContainerEnv(object):
    """Python object representation of devcontainer containerEnv property."""

    def __init__(self, environment_variables: dict):
        self._environment_variables = environment_variables

    def to_dict(self) -> dict:
        return {"containerEnv": self._environment_variables}


class Mounts(object):
    """Python object representation of devcontainer mounts property."""

    def __init__(self, mounts: list):
        self._mounts = mounts

    def to_dict(self) -> dict:
        return {"mounts": self._mounts}


class Name(object):
    """Python object representation of devcontainer name property."""

    def __init__(self, name: str):
        self._name = name

    def to_dict(self) -> dict:
        return {"name": self._name}


class ForwardPorts(object):
    """Python object representation of devcontainer name property."""

    def __init__(self, port: int):
        self._port = port

    def to_dict(self) -> dict:
        return {"forwardPorts": [self._port]}


class AppPort(object):
    """Python object representation of devcontainer name property."""

    def __init__(self, port: int):
        self._port = port

    def to_dict(self) -> dict:
        return {"appPort": [self._port]}


class RunArgs(object):
    """Python object representation of devcontainer runArgs property."""

    def __init__(self, name: str = None, labels: list = None):
        labels = labels or []
        self._run_args = labels
        if name:
            self._run_args.append(f"--name={name}")

    def to_dict(self) -> dict:
        return {"runArgs": self._run_args}


class OverrideCommand(object):
    def __init__(self, override:bool = False):
        self._override = override

    def to_dict(self) -> dict:
        return {"overrideCommand": self._override}


class Extensions(object):
    def __init__(self, extensions:list = None):
        extensions = extensions or []
        self._extensions = extensions

    def to_dict(self) -> dict:
        return {"extensions": self._extensions}


class Settings(object):
    def __init__(self, settings:dict = {}):
        self._settings = settings

    def to_dict(self) -> dict:
        return {"settings": self._settings}

class PostCreateCommand(object):
    def __init__(self, command:str = None):
        self._command = command

    def to_dict(self) -> dict:
        return {"postCreateCommand": self._command}

class PostStartCommand(object):
    def __init__(self, command:str = None):
        self._command = command

    def to_dict(self) -> dict:
        return {"postStartCommand": self._command}



from __future__ import annotations
from sqlalchemy import Column, String, DateTime
from sqlalchemy.sql import func
from sqlalchemy.ext.declarative import declarative_base

Base = declarative_base()

from .db_query import query

class ExperimentMetadata(Base):
    __tablename__ = 'experiment_metadata'
    experiment_id = Column(String,  primary_key = True)
    experiment_name = Column(String)
    created_by = Column(String)
    created_time = Column(DateTime, server_default=func.now())
    last_run_time = Column(DateTime)
    last_runid = Column(String)
    job_type = Column(String)
    def __repr__(self):
        return ("ExperimentMetadata(experiment_id=" + str(self.experiment_id) 
        + ", experiment_name=" + str(self.experiment_name) 
        + ", created_by=" + str(self.created_by) 
        + ", created_time=" + str(self.created_time) 
        + ", last_run_time=" + str(self.last_run_time) 
        + ", last_runid=" + str(self.last_runid ) 
        + ", job_type=" + str(self.job_type) + ")"
        )
    def __eq__(self, other: ExperimentMetadata):
        return self.experiment_name == other.experiment_name and self.experiment_id == other.experiment_id and self.created_by == other.created_by and self.created_time == other.created_time and self.last_run_time == other.last_run_time and self.last_runid == other.last_runid and self.job_type == other.job_type

class ExperimentMetadataDBOperation():
    def __init__(self, Session):
        self.Session = Session

    def select(self, index_entities_request) -> list:
        with self.Session(expire_on_commit=False) as session:
            return query(ExperimentMetadata, index_entities_request, session)

    def insert(self, insert_data):
        data = insert_data
        with self.Session(expire_on_commit=False) as session:
            session.add(insert_data)
            session.commit()
        return data

    def update(self, update_data):
        with self.Session(expire_on_commit=False) as session:
            ret = session.merge(update_data)
            session.commit()
            return ret

    def delete(self, delete_data_experiment_id):
        with self.Session(expire_on_commit=False) as session:
            session.query(ExperimentMetadata).filter(ExperimentMetadata.experiment_id == delete_data_experiment_id).delete()
            session.commit()
            return True


import os
import json
import shutil
from pathlib import Path
import jsonpatch
from .base_debugger import BaseDebugger, Language
from .devcontainer_properties import (
    Settings, 
    Build, 
    Image, 
    Extensions, 
    PostStartCommand, 
    ContainerEnv, 
    OverrideCommand, 
    AppPort, 
    RunArgs, 
    ForwardPorts, 
    PostCreateCommand
)

class RDebugger(BaseDebugger):
    def __init__(self, code_dir: str = None, mounts:dict = None, command:str = None, image:str = None, envs:dict = None) -> None:
        self._command = command
        self._image = image
        self._envs = envs
        super().__init__(code_dir, mounts)

    @property
    def name(self) -> str:
        return "R (Community)"

    @property
    def programming_lang(self) -> Language:
        return Language.R

    def get_devcontainer_settings(self) -> Settings:
        # dev container settings
        # "r.plot.useHttpgd": true,
        # "r.rpath.linux": "",
        return Settings(settings={
            "r.rterm.linux": "/usr/local/bin/radian",
            "r.bracketedPaste": True,
            "[r]": {
                "editor.wordSeparators": "`~!@#%$^&*()-=+[{]}\\|;:'\",<>/?"
            }
        })
    
    def get_devcontainer_extensions(self) -> Extensions:
        return Extensions(extensions=["reditorsupport.r","rdebugger.r-debugger"])

    def get_devcontainer_postcreatecmd(self) -> PostCreateCommand:
        return None

    def get_devcontainer_poststartcmd(self) -> PostStartCommand:
        return PostStartCommand("R")

    def get_devcontainer_envs(self) -> ContainerEnv:
        return ContainerEnv(environment_variables=self._envs) if self._envs else None

    def get_devcontainer_build(self) -> Build:
        return Build(dockerfile_path="Dockerfile",args={"BASE_IMAGE": self._image})

    def get_devcontainer_overridecommand(self) -> OverrideCommand:
        return None

    def get_devcontainer_runargs(self) -> RunArgs:
        return None

    def get_devcontainer_image(self) -> Image:
        return None

    def get_devcontainer_appport(self) -> AppPort:
        return None

    def get_devcontainer_forwardports(self) -> ForwardPorts:
        return None

    def prepare_debug_launch_config(self, debug_dir:str):
        src_launchfile_path = str(Path(Path(__file__).parent, "language", self.programming_lang.value, "launch.json").resolve())
        dest_launchfile_path = str(Path(debug_dir, ".vscode", "launch.json").resolve())
        # delete old launch.json if exist
        if os.path.exists(dest_launchfile_path):
            os.remove(dest_launchfile_path)

        # patch(or replace???) the launch debug mode if the start command is known pattern
        if self._command.startswith("Rscript"):
            launch_mode_config = {
                "type": "R-Debugger",
                "name": "Azure ML: R launch Debug",
                "request": "launch",
                "debugMode": "file",
                "workingDirectory": "${workspaceFolder}", 
            }
            # generate start file path
            command_array = self._command.split(" ")
            cleaned_command_array=[x for x in command_array if x]
            start_file = cleaned_command_array[1]
            launch_mode_config.update({"file": "${workspaceFolder}/src/"+start_file})
            # generate launch commandLineArgs if has extra parameters
            if len(cleaned_command_array) > 2:
                args = ["--args"]
                for x in cleaned_command_array[2:]:
                    args.append(x)
                launch_mode_config.update({"commandLineArgs":args})

            with open(src_launchfile_path) as f:
                data = json.load(f)
            patch = jsonpatch.JsonPatch([ {'op': 'replace', 'path': '/configurations/0', 'value': launch_mode_config}])
            updated_data = patch.apply(data)
            with open(dest_launchfile_path, "w") as f:
                f.write(f"{json.dumps(updated_data, indent=4)}\n")
        else:
            shutil.copyfile(src_launchfile_path, dest_launchfile_path)
            



import os
import json
import shutil
from abc import ABC, abstractmethod
from pathlib import Path
from typing import Optional


from enum import Enum
from .devcontainer_properties import (
    Settings, 
    Build, 
    Image, 
    Name, 
    Extensions, 
    Mounts, 
    PostStartCommand, 
    ContainerEnv, 
    OverrideCommand, 
    AppPort, 
    RunArgs, 
    ForwardPorts, 
    PostCreateCommand
)

class Language(Enum):
    NONE = "none"
    PYTHON = "python"
    R = "R"

class BaseDebugger(ABC):
    def __init__(self, code_dir:str = None, mounts:dict = None) -> None:
        self._properties: Optional[dict] = {}
        self._code_dir = code_dir
        self._mounts = self._reformat_mounts(mounts) if mounts else mounts
        # add code dir mount
        self._mounts.append(f"source={code_dir},target=/src,type=bind")

    @property
    @abstractmethod
    def name(self) -> str:
        pass

    @property
    @abstractmethod
    def programming_lang(self) -> Language:
        pass

    @abstractmethod
    def get_devcontainer_settings(self) -> Settings:
        # dev container settings
        pass
    
    @abstractmethod
    def get_devcontainer_extensions(self) -> Extensions:
        pass

    def get_devcontainer_mounts(self) -> Mounts:
        return Mounts(self._mounts)

    @abstractmethod
    def get_devcontainer_postcreatecmd(self) -> PostCreateCommand:
        pass

    @abstractmethod
    def get_devcontainer_poststartcmd(self) -> PostStartCommand:
        pass

    @abstractmethod
    def get_devcontainer_envs(self) -> ContainerEnv:
        pass

    @abstractmethod
    def get_devcontainer_build(self) -> Build:
        pass

    @abstractmethod
    def get_devcontainer_overridecommand(self) -> OverrideCommand:
        pass

    @abstractmethod
    def get_devcontainer_runargs(self) -> RunArgs:
        pass

    @abstractmethod
    def get_devcontainer_image(self) -> Image:
        pass

    @abstractmethod
    def get_devcontainer_appport(self) -> AppPort:
        pass

    @abstractmethod
    def get_devcontainer_forwardports(self) -> ForwardPorts:
        pass

    def _reformat_mounts(self, mounts: dict) -> list:
        """Reformat mounts from Docker format to DevContainer format.

        :param mounts: Dictionary with mount information for Docker container. For example,
            {
                "<local_source>": {
                    "<mount type i.e. bind>": "<container_dest>"
                }
            }
        :type mounts: dict
        :returns dict: "mounts": ["source=${localWorkspaceFolder}/app-scripts,target=/usr/local/share/app-scripts,type=bind,consistency=cached"]
        """
        devcontainer_mounts = []
        for source, dest in mounts.items():
            for mount_type, container_dest in dest.items():
                devcontainer_mounts.append(f"source={source},target={container_dest},type={mount_type}")
        return devcontainer_mounts

    def _build_devcontainer_properties(self) -> Optional[dict]:
        self._properties = Name(self.name).to_dict()
        
        image = self.get_devcontainer_image()
        build = self.get_devcontainer_build()
        if image:
            self._properties.update(image.to_dict())
        elif build:
            self._properties.update(build.to_dict())
        else:
            msg = "Must provide image or build context for devcontainer.json"
            raise Exception(err=msg)
        
        settings = self.get_devcontainer_settings()
        if settings:
            self._properties.update(settings.to_dict())
        
        extensions = self.get_devcontainer_extensions()
        if extensions:
            self._properties.update(extensions.to_dict())
        
        mounts = self.get_devcontainer_mounts()
        if mounts:
            self._properties.update(mounts.to_dict())

        envs = self.get_devcontainer_envs()
        if envs:
            self._properties.update(envs.to_dict())

        overridecmd = self.get_devcontainer_overridecommand()
        if overridecmd:
            self._properties.update(overridecmd.to_dict())

        appport = self.get_devcontainer_appport()
        if appport:
            self._properties.update(appport.to_dict())

        forwardports = self.get_devcontainer_forwardports()
        if forwardports:
            self._properties.update(forwardports.to_dict())

        postcreatecmd = self.get_devcontainer_postcreatecmd()
        if postcreatecmd:
            self._properties.update(postcreatecmd.to_dict())

        poststartcmd = self.get_devcontainer_poststartcmd()
        if poststartcmd:
            self._properties.update(poststartcmd.to_dict())

        runargs = self.get_devcontainer_runargs()
        if runargs:
            self._properties.update(runargs.to_dict())

    def prepare_devcontainer_config(self, debug_dir:str):
        self._build_devcontainer_properties()
        devcontainer_config_path = str(Path(debug_dir, ".devcontainer", "devcontainer.json").resolve())
        # delete if old on exist
        if os.path.exists(devcontainer_config_path):
            os.remove(devcontainer_config_path)
        with open(devcontainer_config_path, "w") as f:
            f.write(f"{json.dumps(self._properties, indent=4)}\n")

    def prepare_devcontainer_dockerfile(self, debug_dir:str):
        src_dockerfile_path = str(Path(Path(__file__).parent, "language", self.programming_lang.value, "Dockerfile").resolve())
        dest_dockerfile_path = str(Path(debug_dir, ".devcontainer", "Dockerfile").resolve())
        # delete dest if exist
        if os.path.exists(dest_dockerfile_path):
            os.remove(dest_dockerfile_path)
        shutil.copyfile(src_dockerfile_path, dest_dockerfile_path)

    def prepare_debug_launch_config(self, debug_dir:str):
        """prepare vscode debug launch.json file (if needed), by default this just copy the predefined launch.json file, each language debugger
        can customize this with their own implementation.

        :param debug_dir: local debug root dir which will be loaded in the devcontainer
        """
        src_launchfile_path = str(Path(Path(__file__).parent, "language", self.programming_lang.value, "launch.json").resolve())
        dest_launchfile_path = str(Path(debug_dir, ".vscode", "launch.json").resolve())
        # delete old launch.json if exist
        if os.path.exists(dest_launchfile_path):
            os.remove(dest_launchfile_path)
        shutil.copyfile(src_launchfile_path, dest_launchfile_path)

    def prepare_debug_data(self, debug_dir:str):
        # debug_dir should be the path in container here, while calculating the devcontainer invoke command we should use the real path on the node.
        # source code mount from host and copied to debug_dir
        devcontainer_path = Path(debug_dir, ".devcontainer")
        vscode_path = Path(debug_dir, ".vscode")
        devcontainer_path.mkdir(parents=True, exist_ok=True)
        vscode_path.mkdir(parents=True, exist_ok=True)
        self.prepare_devcontainer_config(debug_dir)
        self.prepare_devcontainer_dockerfile(debug_dir)
        self.prepare_debug_launch_config(debug_dir)
        


        



from .graph import Graph, GraphEdge, GraphDatasetNode, DataSetDefinition, PortInfo
from .component import CommandComponent, ComponentType
from .input_output import GeneralInput, GeneralOutput, InputType,OutputType, InputOutputModes, RunInputOutputBase
from .index_entities_request import IndexEntitiesRequest, IndexEntitiesRequestFilter, IndexEntitiesRequestOrder, IndexEntitiesRequestFilterSchema, IndexEntitiesRequestOrderSchema, IndexEntitiesRequestSchema
from .db_setup import db_setup, drop_tables, DEFAULT_DB_PATH
from .constants import RunType


from enum import Enum
from typing import Dict
import uuid
from datetime import datetime
from sqlalchemy.exc import SQLAlchemyError
from piprundb import RunHistory, RunHistoryDBOperation, ComponentMetadata, ComponentMetadataDBOperation, ExperimentMetadata, ExperimentMetadataDBOperation
from pipruncommon import GeneralInput, GeneralOutput
from pipruncommon import InputType, OutputType, InputOutputModes
from .constants import JobRunningStatus
import getpass

from .graph import PipelineExecutionGraph, PipelineExecutionNode
from pipruncommon import Graph, GraphEdge, GraphDatasetNode, DataSetDefinition, PortInfo
from pipruncommon import RunInputOutputBase
from pipruncommon import CommandComponent as LocalCommandComponent
from azure.ai.ml import Input,Output
from azure.ai.ml.constants import JobType
from azure.ai.ml.entities import Component, CommandComponent
from azure.ai.ml.entities._job.pipeline._io import NodeInput, NodeOutput, PipelineInput, PipelineOutput

class MetadataUpdateType(Enum):
    READY_TIME= 0,
    START_TIME = 1,
    FINISH_TIME = 2

class MetadataRecorder(object):
    def __init__(self, Session):
        self.Session = Session
        self.run_history_db_operation = RunHistoryDBOperation(Session)
        self.component_metadata_db_operation = ComponentMetadataDBOperation(Session)
        self.experiment_metadata_db_operation = ExperimentMetadataDBOperation(Session)
        self.metadata_generated = False

    def add_pipeline_run_metadata(self, pipeline: PipelineExecutionGraph):
        # create graph
        graph = MetadataRecorder._generate_run_graph(self, pipeline)
        # create inputs/outputs
        inputs = []
        outputs = []
        parameters = []
        for k, v in pipeline._job.inputs.items():
            if isinstance(v._data, (Input, PipelineInput, Output, NodeOutput)):
                # non-parameter type input
                inputs.append(RunInputOutputBase(name=k, type=v._data.type, uri = v._data.path, mode = v._data.mode))
            else:
                # parameter
                parameters.append(RunInputOutputBase(name=k, type=v.type, value=v._data))

        for k, v in pipeline._job.outputs.items():
            output_type = v.type if v.type is not None else OutputType.URI_FOLDER
            output_mode = v.mode if v.mode is not None else InputOutputModes.RW_MOUNT
            outputs.append(RunInputOutputBase(name=k, type=output_type, mode=output_mode))
        # create jobs info
        jobs = {}
        for name, job in pipeline._job.jobs.items():
            # if isinstance(job, BaseNode):
            rest_node_dict = job._to_rest_object()
            jobs[name] = rest_node_dict

        cur_time = datetime.utcnow()

        run_history = RunHistory(run_id=pipeline._run_id,
            run_number=1,
            display_name = pipeline._name,
            run_type="azureml.PipelineRun",
            job_type=pipeline._run_type,
            run_mode=pipeline._run_mode,
            created_by=getpass.getuser(),
            created_time=cur_time,
            ready_time=cur_time, # this might need to change in future if we allow queue in local
            status=pipeline._status,
            total_steps=len(pipeline._nodes),
            log_file=pipeline._log_path, 
            parent_runid=pipeline._parent_run_id,
            root_runid=pipeline._root_run_id,
            node_id=pipeline._node_id,
            inputs=inputs,
            outputs=outputs,
            experiment_id = pipeline._experiment_id,
            tags = pipeline._job.tags,
            jobs = jobs,
            run_root_dir=pipeline._output_root_dir,
            parameters=parameters,
            graph=graph)
        self.run_history_db_operation.insert(run_history)
        self.metadata_generated = True
        # add or update experiment
        experiment_metadata = ExperimentMetadata(
            experiment_id = pipeline._experiment_id,
            experiment_name = pipeline._experiment_name,
            created_by = getpass.getuser(),
            last_run_time = cur_time,
            last_runid = pipeline._run_id,
            job_type = pipeline._job.type
        )
        self.experiment_metadata_db_operation.update(experiment_metadata)

        # add job run
        for n, node in pipeline._nodes.items():
            MetadataRecorder.add_job_run_metadata(self, node)

    def add_job_run_metadata(self, job: PipelineExecutionNode):
        component_str = job._job._component._to_yaml()
        component_id = job._component_id
        # build job inputs/outputs
        inputs = []
        outputs = []
        parameters = []
        for k,v in job._inputs.items():
            pip_reference = None
            param_data = None
            has_path = False
            path_type = None
            if isinstance(v._data, PipelineInput):
                pip_reference = v._data._name
                if isinstance(v._data._data, Input):
                    has_path = True
                    path_type = v._data._data.type
                else:
                    param_data = v._data
            elif isinstance(v._data, (Input, Output, NodeOutput)):
                has_path = True
                path_type = v._data.type
            else:
                param_data = v
            
            if has_path:
                input_mode = v._data.mode
                if input_mode is None:
                    input_mode = InputOutputModes.RO_MOUNT
                inputs.append(RunInputOutputBase(name=k, type=path_type, uri = v._data.path, mode = v._data.mode, pipelineReference=pip_reference))
            else:
                parameters.append(RunInputOutputBase(name=k, type=param_data.type, value=param_data._data, pipelineReference=pip_reference))

        for k, v in job._outputs.items():
            pip_reference = None
            if isinstance(v._data, PipelineOutput):
                pip_reference = v._data._name
                outputs.append(RunInputOutputBase(name=k, type=v._data.type, uri = v._data.path, mode = v._data.mode, pipelineReference=pip_reference))
            elif isinstance(v._data, Output):
                outputs.append(RunInputOutputBase(name=k, type=v._data.type, uri = v._data.path, mode = v._data.mode))

        # job run record
        run_history = RunHistory(run_id=job._run_id,
            run_number=1,
            display_name = job._job.name,
            run_type="azureml.StepRun",
            job_type=job._run_type,
            run_mode=job._curated_env_default_run_mode,
            created_by=getpass.getuser(),
            created_time=datetime.utcnow(),
            status=job._status,
            total_steps=1,  # need to change in future when subgraph is supported
            log_file=job._log_path, 
            parent_runid=job._parent_run_id,
            root_runid=job._root_run_id,
            node_id=job._node_id,
            component_id = component_id,
            experiment_id = job._experiment_id,
            component_base_path = str(job._job._component.base_path),
            run_root_dir=job._outputdir,
            inputs=inputs,
            outputs=outputs,
            parameters=parameters,
            tags = job._job.tags)
        self.run_history_db_operation.insert(run_history)
        # component info
        exist = self.component_metadata_db_operation.get(component_id)
        if exist is None:
            # need to add new component record
            component_type = job._job._component.type
            if component_type == JobType.COMMAND:
                originComponent: CommandComponent = job._job._component
                # build component inputs/outputs
                component_inputs: Dict[str, GeneralInput] = {}
                component_outputs: Dict[str, GeneralOutput] = {}
                for k, v in originComponent.inputs.items():
                    if isinstance(v, Input):
                        path = v.get("path")
                        default = v.default
                        min = v.get("min")
                        max = v.get("max")
                        component_inputs[k] = GeneralInput(name=k, type=v.type, path = path, default = default, min = min, max=max, optional=v.optional, mode=v.mode)
                for k, v in originComponent.outputs.items():
                    if isinstance(v, Output):
                        path = v.get("path")
                        component_outputs[k] = GeneralOutput(name=k, type=v.type, path=path, mode=v.mode)               
                component = LocalCommandComponent(component_id=component_id,
                    component_type=component_type,
                    description= originComponent.description,
                    name = originComponent.name,
                    display_name=originComponent.display_name,
                    version=originComponent.version,
                    tags=originComponent.tags,
                    properties=originComponent.properties,
                    is_deterministic=originComponent.is_deterministic,
                    command=originComponent.command,
                    code=originComponent.code,
                    inputs=component_inputs,
                    outputs=component_outputs,
                    raw_content=component_str
                    )
                component_meta = ComponentMetadata(component_id=component_id, component_type = component_type, created_time=datetime.utcnow(), component = component)
                self.component_metadata_db_operation.insert(component_meta)
            else:
                raise Exception(f"component type {component_type} not supported yet!")
        else:
            print("component already exist!")

    def update_job_status(self, run_id:str, status:JobRunningStatus, update_type: MetadataUpdateType):
        if update_type == MetadataUpdateType.READY_TIME:
            self.run_history_db_operation.update(RunHistory(run_id=run_id, status=status, ready_time=datetime.utcnow()))
        elif update_type == MetadataUpdateType.START_TIME:
            self.run_history_db_operation.update(RunHistory(run_id=run_id, status=status, start_time=datetime.utcnow()))
        elif update_type == MetadataUpdateType.FINISH_TIME:
            finish_time = datetime.utcnow()
            record = self.run_history_db_operation.get(run_id)
            duration = finish_time - record.start_time      # ? decrease 1970-01-01
            self.run_history_db_operation.update(RunHistory(run_id=run_id, status=status, finish_time=finish_time, duration = duration))

    def update_container_build_info(self, component_id:str, docker_file_content:str):
        if docker_file_content is not None:
            self.component_metadata_db_operation.update(ComponentMetadata(component_id=component_id, docker_file_content=docker_file_content))

    def update_container_run_info(self, run_id:str, container_image_id:str, run_command:str, mounts:dict, code_dir:str):
        self.run_history_db_operation.update(RunHistory(
            run_id=run_id, 
            container_image=container_image_id, 
            run_command=run_command,
            mounts=mounts,
            code_dir=code_dir))

    def update_container_id(self, run_id:str, container_id:str):
        self.run_history_db_operation.update(RunHistory(run_id=run_id, container_id=container_id))

    def _generate_run_graph(self, pipeline_graph: PipelineExecutionGraph) -> Graph:
        datasets: Dict[str, GraphDatasetNode] = {}
        edges = []
        for k, v in pipeline_graph._pipeline_inputs.items():
            if isinstance(v._data, (Input, PipelineInput, Output, NodeOutput)):
                # data source input
                node_id = str(uuid.uuid1())
                ds = GraphDatasetNode(id = node_id, dataSetDefinition=DataSetDefinition(parameterName=k, dataTypeShortName="path"))
                datasets[k] = ds

        for name, node in pipeline_graph._nodes.items():
            dst_node_id = node._node_id
            for k, v in node._inputs.items(): 
                if isinstance(v, NodeInput):
                    dst_port = PortInfo(nodeId=dst_node_id, portName=k)
                    if isinstance(v._data, NodeOutput):
                        # input is a job output
                        src_job_name = v._data._owner.name
                        src_node_id = pipeline_graph._nodes[src_job_name]._node_id
                        src_port_name = v._data._name
                        src_port = PortInfo(nodeId=src_node_id, portName=src_port_name)
                        edges.append(GraphEdge(sourceOutputPort=src_port, destinationInputPort=dst_port))
                    elif isinstance(v._data, PipelineInput) and hasattr(v._data._data, "path"):
                        # input is a data source
                        graph_port_name = v._data._name
                        src_dataset :GraphDatasetNode= datasets.get(graph_port_name)
                        src_node_id = src_dataset.id
                        src_port = PortInfo(nodeId=src_node_id)
                        edges.append(GraphEdge(sourceOutputPort=src_port, destinationInputPort=dst_port))
            
            for k, v in node._job.outputs.items():
                # based on output reference to detect whether this is pipeline output
                if isinstance(v, NodeOutput) and isinstance(v._data, PipelineOutput):
                    src_port = PortInfo(nodeId=dst_node_id, portName=k)
                    dst_port = PortInfo(graphPortName=v._data._name)
                    edges.append(GraphEdge(sourceOutputPort=src_port, destinationInputPort=dst_port))
        ds = [p for _, p in datasets.items()]
        return Graph(datasetNodes=ds, edges=edges)


import subprocess
import os
import re
import sys
import logging
import logging.handlers
from typing import Dict
from pathlib import Path

from azure.ai.ml import MLClient, Input, Output
from azure.ai.ml.entities import PipelineJob, CommandJob, Job
from azure.ai.ml.entities._builders import Command, Sweep
from azure.ai.ml.entities._job.pipeline._io import NodeOutput, PipelineInput, PipelineOutput
from azure.ai.ml.constants import AssetTypes

import docker
from docker import DockerClient
from docker.types import Mount
from docker.errors import NotFound, APIError
from docker.models.containers import Container

from .graph import PipelineExecutionNode, PipelineExecutionGraph
from .image_builder import LocalImageBuilder
from .constants import JobRunningStatus, LocalRunMode, CuratedEnvironmentDefaultRunMode, JobLocalRunMode, BLOB_WASBS_REGEX, BLOB_HTTPS_PATTERN
from .metadata import MetadataRecorder, MetadataUpdateType
from ._utils import server_ok, download
from .logger import log_setup, get_logger

import uuid
from piprundb import all_tables
from pipruncommon import db_setup, DEFAULT_DB_PATH

# def get_logger(logger_name, log_level:int = logging.INFO, log_path:str = None, std_out:bool = False, log_formatter:logging.Formatter = None) -> logging.Logger:
#     logger = logging.getLogger(logger_name)
#     logger.setLevel(log_level)
#     if log_path:
#         fh = logging.handlers.RotatingFileHandler(log_path, mode='w', backupCount=5)
#         if log_formatter:
#             fh.setFormatter(log_formatter)
#         logger.addHandler(fh)
#     if std_out:
#         logger.addHandler(logging.StreamHandler(sys.stdout))
#     return logger

# def get_infra_logger(logger_name, log_path:str = None, ) -> logging.Logger:
#     formatter = logging.Formatter(fmt='%(asctime)s %(name)-12s %(levelname)-8s %(message)s',
#                                     datefmt='%m-%d-%y %H:%M:%S')
#     return get_logger(logger_name=logger_name, log_level=logging.DEBUG, log_path=log_path, std_out=True, log_formatter=formatter)

# def log_setup(log_path):
#     return get_infra_logger("orchestrator",log_path)
HOME_PATH = os.path.join(os.path.sep, str(Path.home()), ".azureml", "piprun")

def run(
    *,
    output_root_dir: str=None,
    job: Job,
    experiment_name: str = None,
    ml_client: MLClient = None,
    run_mode: LocalRunMode = LocalRunMode.COMPUTE_LOCAL_METRICS_LOCAL,
    curated_env_default_run_mode = CuratedEnvironmentDefaultRunMode.CONTAINER,
     **kwargs,
):  
    """
    """
    run_id = str(uuid.uuid1())
    # currently only support COMPUTE_LOCAL_METRICS_LOCAL
    if run_mode != LocalRunMode.COMPUTE_LOCAL_METRICS_LOCAL:
        raise Exception(f"found unsupported mode: {run_mode}, only mode {LocalRunMode.COMPUTE_LOCAL_METRICS_LOCAL} is supported!")
    if isinstance(job, CommandJob):
        raise Exception("standalone job run not supported yet!")
    elif isinstance(job, PipelineJob):
        db_dir = os.path.dirname(DEFAULT_DB_PATH)
        if not os.path.exists(db_dir):
            os.makedirs(db_dir)
        Session = db_setup(DEFAULT_DB_PATH, all_tables)
        metadata_recorder = MetadataRecorder(Session)
        try:
            _run_pipeline(output_root_dir=output_root_dir, 
                pipeline_job=job, 
                experiment_name=experiment_name, 
                metadata_recorder = metadata_recorder, 
                ml_client=ml_client, run_mode=run_mode, 
                curated_env_default_run_mode=curated_env_default_run_mode,
                run_id=run_id)
        except:
            if metadata_recorder.metadata_generated:
                metadata_recorder.update_job_status(run_id, JobRunningStatus.FAIL, MetadataUpdateType.FINISH_TIME)
            raise sys.exc_info()[0]
        finally:
            Session.close_all()
    else:
        raise Exception(f"found unsupported Job type: {run_mode}, only mode {LocalRunMode.COMPUTE_LOCAL_METRICS_LOCAL} is supported!")


def _run_pipeline(
    *,
    output_root_dir: str=None,
    pipeline_job: PipelineJob,
    experiment_name: str = None,
    metadata_recorder: MetadataRecorder = None,
    ml_client: MLClient = None,
    run_mode: LocalRunMode = LocalRunMode.COMPUTE_LOCAL_METRICS_LOCAL,
    curated_env_default_run_mode = CuratedEnvironmentDefaultRunMode.CONTAINER,
     **kwargs,
):
    """
    """
    # currently only support COMPUTE_LOCAL_METRICS_LOCAL
    if run_mode != LocalRunMode.COMPUTE_LOCAL_METRICS_LOCAL:
        raise Exception(f"found unsupported mode: {run_mode}, only mode {LocalRunMode.COMPUTE_LOCAL_METRICS_LOCAL} is supported!")
    # validate pipeline job, currently only support command job
    for _, job_instance in pipeline_job.jobs.items():
        if not isinstance(job_instance, Command): #check sweep job?
            raise Exception(f"Not supported pipeline job type: {type(job_instance)}")
    
    if not os.path.exists(HOME_PATH):
        os.makedirs(HOME_PATH)
    if experiment_name is None:
        # use Default as experiment name if not provided
        experiment_name = "Default"
    _run_id = kwargs.get("run_id", str(uuid.uuid1()))
    if output_root_dir is None:
        output_root_dir = HOME_PATH #os.getcwd()
    if not os.path.isabs(output_root_dir):
        output_root_dir = os.path.join(os.path.sep, HOME_PATH, output_root_dir)
    log_path = os.path.join(os.path.sep, output_root_dir, _run_id, "orchestrator.log")
    log_dir = os.path.dirname(log_path)
    if not os.path.exists(log_dir):
        os.makedirs(log_dir)
    log = log_setup(log_path=log_path)
    if pipeline_job.name is None:
        pipeline_job.name = _run_id
    # build pipeline execution graph
    graph = PipelineExecutionGraph(output_root_dir=output_root_dir, pipeline_job=pipeline_job, experiment_name=experiment_name, curated_env_default_run_mode=curated_env_default_run_mode, run_id = _run_id, log_path=log_path)
    # record pipeline run metadata
    if metadata_recorder is None:
        Session = db_setup(DEFAULT_DB_PATH, all_tables)
        metadata_recorder = MetadataRecorder(Session)
    metadata_recorder.add_pipeline_run_metadata(graph)
    metadata_recorder.update_job_status(graph._run_id, JobRunningStatus.RUNNING, MetadataUpdateType.START_TIME)
    image_builder = LocalImageBuilder(ml_client, log_path)
    while not graph.is_finish():
        nodes_to_run = graph.get_ready_nodes()
        for name, node in nodes_to_run.items():
            metadata_recorder.update_job_status(node._run_id, JobRunningStatus.NOT_STARTED, MetadataUpdateType.READY_TIME)
        for name, node in nodes_to_run.items():
            # better to record image build time separately
            metadata_recorder.update_job_status(node._run_id, JobRunningStatus.RUNNING, MetadataUpdateType.START_TIME)
            run_mode = node.get_run_mode()
            success = True
            try:
                if run_mode == JobLocalRunMode.NATIVE:
                    # run as native process
                    success = _run_with_native_process(node, log)
                elif run_mode == JobLocalRunMode.DOCKER:
                    # build image using job component environment
                    # TODO: check whether need to build image first
                    image_name, docker_file_content = image_builder.build_container_image(pipeline_job.name, node._job)
                    metadata_recorder.update_container_build_info(node._component_id, docker_file_content)
                    success = _run_with_docker_image(image_name, node, log, metadata_recorder)
                else:
                    # TODO: define error type for this
                    raise Exception(f"Found unsupported run mode: {run_mode}")
            except:
                log.error("Oops! %s occurred.", sys.exc_info()[0])
                success = False
                metadata_recorder.update_job_status(node._run_id, JobRunningStatus.FAIL, MetadataUpdateType.FINISH_TIME)
                raise sys.exc_info()[0]

            if not success:
                log.info(f"pipeline {pipeline_job.name} stopped due to failed job {name}!")
                metadata_recorder.update_job_status(node._run_id, JobRunningStatus.FAIL, MetadataUpdateType.FINISH_TIME)
                metadata_recorder.update_job_status(graph._run_id, JobRunningStatus.FAIL, MetadataUpdateType.FINISH_TIME)
                return
            else:
                metadata_recorder.update_job_status(node._run_id, JobRunningStatus.SUCCESS, MetadataUpdateType.FINISH_TIME)

    log.info(f"pipeline {pipeline_job.name} finished successfully!")
    metadata_recorder.update_job_status(graph._run_id, JobRunningStatus.SUCCESS, MetadataUpdateType.FINISH_TIME)

def _run_with_native_process(node: PipelineExecutionNode, log: logging.Logger) -> bool:
    # run as native process
    name = node._job.name
    working_dir = _get_working_dir(node)
    run_command = node.command()
    cmd_array = run_command.split()
    log.info(f"############################################{name}##############################################")
    log.debug(f"start running job {name}")
    params = {}
    params["capture_output"]=True
    if working_dir is not None:
        params["cwd"] = working_dir
    process = subprocess.run(cmd_array, **params)
    success = (process.returncode == 0)
    job_logger = get_logger(node._job.name, log_path=node._log_path, std_out=True)
    # TODO: update to streaming mode
    print_pretty(process.stdout.decode("utf-8"), job_logger)
    node.mark_run_result(success)
    log.info(f"job {name} finished with result: {success}")
    if not success:
        err_data = process.stderr.decode("utf-8")
        log.info(f"Run job {name} with failure return code: {process.returncode}, error: {err_data}")
        print_pretty(process.stderr.decode("utf-8"), log)
    return success

def _run_with_docker_image(docker_image_name:str, node: PipelineExecutionNode, log: logging.Logger, metadata_recorder: MetadataRecorder, docker_client: DockerClient=None) -> bool:
    container_name= node._job.name
    log.info(f"##########################################{container_name} start################################################")
    log.debug(f"docker image: {docker_image_name}")
    if docker_client is None:
        docker_client = docker.from_env()
    # get working dir
    working_dir = _get_working_dir(node)
    log.debug(f"working dir: {working_dir}")
    # build mount for input/output path
    volumn_mapping: Dict[str, str] = {}
    template_params = {}
    idx = 1
    # TODO: do we need to care about mounting mode? 
    # currently by default ro mode for input and rw mode for output
    for input_name, input in node.inputs().items():
        key = f"inputs_{input_name}"
        has_path = False
        data = None
        path_type = None
        if isinstance(input._data, PipelineInput):
            if isinstance(input._data._data, Input):
                has_path = True
                path_type = input._data._data.type
            else:
                data = input._data._data
            pass
        elif isinstance(input._data, (Input, Output, NodeOutput)):
            has_path = True
            path_type = input._data.type
        else:
            data = input._data

        if has_path:
            if path_type == AssetTypes.URI_FILE:
                input_path = prepare_input_path_for_remote_file(input._data.path, node._inputs_temp_dir, log)
                dir_name = os.path.dirname(input_path)
                file_name = os.path.basename(input_path)
                if dir_name in volumn_mapping.keys():
                    template_params[key] = os.path.join(volumn_mapping[input_path],file_name)
                else:
                    mapping_path = f"/{node._job.name}/inputs/{idx}/{file_name}"
                    idx += 1
                    volumn_mapping[input_path] = mapping_path
                    template_params[key] = mapping_path
            else:
                if input._data.path not in volumn_mapping.keys():
                    mapping_path = f"/{node._job.name}/inputs/{idx}"
                    idx += 1
                    volumn_mapping[input._data.path] = mapping_path
                    template_params[key] = mapping_path
                else:
                    template_params[key] = volumn_mapping[input._data.path]
        else:
            template_params[key] = data
    
    idx = 1
    shared_folder: Dict[str, bool] = {}
    for output_name, output in node.outputs().items():
        key = f"outputs_{output_name}"
        if isinstance(output._data, Output):
            # if output._data.type == AssetTypes.URI_FOLDER or output._data.type == AssetTypes.MLFLOW_MODEL:
            if output._data.type == AssetTypes.URI_FILE:
                dir_name = os.path.dirname(output._data.path)
                file_name = os.path.basename(output._data.path)
                if dir_name in volumn_mapping.keys:
                    # shared folder for input/output
                    shared_folder[dir_name] = True
                    template_params[key] = os.path.join(volumn_mapping[dir_name],file_name)
                else:
                    mapping_path = f"/{node._job.name}/outputs/{idx}/{file_name}"
                    idx += 1
                    volumn_mapping[output._data.path] = mapping_path
                    template_params[key] = mapping_path
            else:
                if output._data.path not in volumn_mapping.keys():
                    mapping_path = f"/{node._job.name}/outputs/{idx}"
                    idx += 1
                    volumn_mapping[output._data.path] = mapping_path
                    template_params[key] = mapping_path
                else:
                    # shared folder for input/output
                    shared_folder[output._data.path] = True
                    template_params[key] = volumn_mapping[output._data.path]
        else:
            template_params[key] = output._data
    log.debug(f'volumn_mapping: {volumn_mapping}')
    log.debug(f'template_params: {template_params}')
    command_template = node._command_template
    command_in_docker = command_template.substitute(template_params)
    log.info(f"Command run in docker: {command_in_docker}")
    # rebuild docker command
    mnts = []
    mnts_meta = {}
    for src, target in volumn_mapping.items():
        read_only = "/inputs/" in target and src not in shared_folder.keys()
        mnt = Mount(source=src, type='bind', read_only=read_only, target=target)
        mnts.append(mnt)
        mnts_meta.update({src: {"bind": target}})

    metadata_recorder.update_container_run_info(node._run_id, docker_image_name, command_in_docker, mnts_meta, working_dir)

    # add code dir mnting
    working_dir_docker = working_dir
    if working_dir is not None:
        dir_name = os.path.dirname(working_dir)
        base_name = os.path.basename(working_dir)
        log.debug(f"dirname: {dir_name}, basename: {base_name}")
        if dir_name in volumn_mapping.keys(): 
            working_dir_docker = f"{volumn_mapping[dir_name]}/{base_name}"
        else:
            working_dir_docker = f"/{node._job.name}/cwd/{base_name}"
            mnt = Mount(source=working_dir, type='bind', target=working_dir_docker)
            mnts.append(mnt)
            log.debug(f"binding {working_dir} to {working_dir_docker} in volumes")
    try:
        container = docker_client.containers.get(container_name)
        if container.status == "running":
            container.stop()
        container.remove()
    except NotFound:
        pass
    params={}
    if working_dir_docker is not None:
        params["working_dir"] = working_dir_docker
    # check whether local mlflow tracking server starts or not, if it start, then set env for container
    mlflow_start = server_ok("http://localhost:5000")
    if mlflow_start:
        params["environment"] = ["MLFLOW_TRACKING_URI=http://host.docker.internal:5000/"]
        mlflow_dir = os.path.join(HOME_PATH, "mlflow")
        target_dir = "/metadata/mlflow"
        mnt = Mount(source=mlflow_dir, type='bind', target=target_dir)
        mnts.append(mnt)
        log.debug(f"binding {mlflow_dir} to {target_dir} in volumes")
    container: Container = docker_client.containers.run(docker_image_name, detach=True, mounts=mnts, name=container_name, command=["/bin/bash", "-c", command_in_docker], **params)
    log.info(f"container-id: {container.id}")
    log_gen = container.logs(stream=True)
    container_logger = get_logger(node._job.name, log_path=node._log_path)
    try:
        while True:
            line = next(log_gen).decode("utf-8").strip()
            if len(line) > 0:
                container_logger.info(line)
    except StopIteration:
        log.info(f'log stream ended for {container_name}')
    result = container.wait()
    success = result['StatusCode'] == 0
    node.mark_run_result(success)
    log.info(f"Container {node._job.name} finished with result: {success}")
    if not success:
        log.info(f"Container {node._job.name} run result: {result}")
    metadata_recorder.update_container_id(node._run_id, container.id)
    log.info(f"##########################################{container_name} Finish################################################")
    return success

def _get_working_dir(node: PipelineExecutionNode) -> str:
    working_dir = None
    if node._job._component.code is not None:
        working_dir = str(node._job._component.code)
    if working_dir is not None and not os.path.isabs(working_dir):
        if  os.path.isabs(node._job._component.base_path):
            working_dir = os.path.join(node._job._component.base_path, working_dir.strip("./"))
        else:
            working_dir = os.path.abspath(working_dir)
    return working_dir

def print_pretty(data: str, log: logging.Logger):
    lines = data.split("\n")
    for l in lines:
        data = l.strip()
        if len(data) > 0:
            log.info(l)
            # print(data)

def prepare_input_path_for_remote_file(path:str, inputs_dir: str, log: logging.Logger) -> str:    
    x = re.match(BLOB_WASBS_REGEX, path)
    if x:
        remote_url = BLOB_HTTPS_PATTERN.format(x.group(2), x.group(1), x.group(3))
        if not os.path.exists(inputs_dir):
            os.makedirs(inputs_dir)
        file_path = os.path.join(inputs_dir, x.group(3))
        if download(remote_url, file_path):
            return file_path
        else:
            log.info(f"Downloading file {remote_url} failed!")
    return path


import sys
import logging
import logging.handlers

def get_logger(logger_name, log_level:int = logging.INFO, log_path:str = None, std_out:bool = False, log_formatter:logging.Formatter = None) -> logging.Logger:
    logger = logging.getLogger(logger_name)
    logger.setLevel(log_level)
    if log_path:
        fh = logging.FileHandler(log_path, mode='a')
        if log_formatter:
            fh.setFormatter(log_formatter)
        logger.addHandler(fh)
    if std_out:
        logger.addHandler(logging.StreamHandler(sys.stdout))
    return logger

def get_infra_logger(logger_name, log_path:str = None, ) -> logging.Logger:
    formatter = logging.Formatter(fmt='%(asctime)s %(name)-12s %(levelname)-8s %(message)s',
                                    datefmt='%m-%d-%y %H:%M:%S')
    return get_logger(logger_name=logger_name, log_level=logging.DEBUG, log_path=log_path, std_out=True, log_formatter=formatter)

def log_setup(log_path):
    return get_infra_logger("orchestrator",log_path)


from .base_debugger import BaseDebugger, Language
from .devcontainer_properties import (
    Settings, 
    Build, 
    Image, 
    Extensions, 
    PostStartCommand, 
    ContainerEnv, 
    OverrideCommand, 
    AppPort, 
    RunArgs, 
    ForwardPorts, 
    PostCreateCommand
)

POST_START_COMMAND_PREFIX = "cd ${containerWorkspaceFolder}/src && python -m debugpy --listen 0.0.0.0:5678 --wait-for-client"

class PythonDebugger(BaseDebugger):
    def __init__(self, code_dir: str = None, mounts:dict = None, command:str = None, image:str = None, envs:dict = None) -> None:
        self._command = command
        self._image = image
        self._envs = envs
        super().__init__(code_dir, mounts)
    
    @property
    def programming_lang(self) -> Language:
        return Language.PYTHON

    @property
    def name(self) -> str:
        return "Python 3"
    
    def get_devcontainer_settings(self) -> Settings:
        # dev container settings
        return Settings(settings={
            "terminal.integrated.shell.linux": "/bin/bash",
            "python.pythonPath": "/usr/local/bin/python",
            "python.linting.enabled": True,
            "python.linting.pylintEnabled": True,
            "python.formatting.autopep8Path": "/usr/local/py-utils/bin/autopep8",
            "python.formatting.blackPath": "/usr/local/py-utils/bin/black",
            "python.formatting.yapfPath": "/usr/local/py-utils/bin/yapf",
            "python.linting.banditPath": "/usr/local/py-utils/bin/bandit",
            "python.linting.flake8Path": "/usr/local/py-utils/bin/flake8",
            "python.linting.mypyPath": "/usr/local/py-utils/bin/mypy",
            "python.linting.pycodestylePath": "/usr/local/py-utils/bin/pycodestyle",
            "python.linting.pydocstylePath": "/usr/local/py-utils/bin/pydocstyle",
            "python.linting.pylintPath": "/usr/local/py-utils/bin/pylint"
        })
    
    def get_devcontainer_extensions(self) -> Extensions:
        return Extensions(extensions=["ms-python.python"])

    def get_devcontainer_postcreatecmd(self) -> PostCreateCommand:
        return PostCreateCommand(command="rm -rf ${containerWorkspaceFolder}/src && ln -s /src ${containerWorkspaceFolder}/src")

    def get_devcontainer_poststartcmd(self) -> PostStartCommand:
        commandprefix = "python"
        designerprefix = "mldesigner"
        if self._command.startswith(commandprefix):
            command = self._command.removeprefix(commandprefix)
            debug_command = POST_START_COMMAND_PREFIX + command
            return PostStartCommand(debug_command)
        # support new version mldesigner debug
        elif self._command.startswith(designerprefix):
            command = self._command.removeprefix(designerprefix)
            debug_command = "designer_path=`which mldesigner` && " + POST_START_COMMAND_PREFIX + " $designer_path " + command
            return PostStartCommand(debug_command)
        else:
            # user need to do some customized work if the command is not run in python
            return None

    def get_devcontainer_envs(self) -> ContainerEnv:
        return ContainerEnv(environment_variables=self._envs) if self._envs else None

    def get_devcontainer_build(self) -> Build:
        return Build(dockerfile_path="Dockerfile",args={"BASE_IMAGE": self._image})

    def get_devcontainer_overridecommand(self) -> OverrideCommand:
        return None

    def get_devcontainer_runargs(self) -> RunArgs:
        return None

    def get_devcontainer_image(self) -> Image:
        return None

    def get_devcontainer_appport(self) -> AppPort:
        return None

    def get_devcontainer_forwardports(self) -> ForwardPorts:
        return None